# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.autodock"
__author__ = "Patrick Kunzmann"
__all__ = ["VinaApp"]

import copy
from tempfile import NamedTemporaryFile
import numpy as np
from ..localapp import LocalApp, cleanup_tempfile
from ..application import AppState, requires_state
from ...structure.io.pdbqt import PDBQTFile
from ...structure.residues import get_residue_starts_for, get_residue_masks
from ...structure.bonds import find_connected
from ...structure.error import BadStructureError


class VinaApp(LocalApp):
    """
    Dock a ligand to a receptor molecule using *AutoDock Vina*.

    Parameters
    ----------
    ligand : AtomArray
        The structure of the receptor molecule.
        Must have an associated :class:`BondList`.
        An associated ``charges`` annotation is recommended for proper
        calculation of partial charges.
    receptor : AtomArray, shape=(n,)
        The structure of the receptor molecule.
        Must have an associated :class:`BondList`.
        An associated ``charges`` annotation is recommended for proper
        calculation of partial charges.
    center : ndarray, shape=(3,), dtype=float
        The *xyz* coordinates for the center of the search space.
    size : ndarray, shape=(3,), dtype=float
        The size of the search space in *xyz* directions.
    flexible : ndarray, shape=(n,), dtype=bool, optional
        A boolean mask that indicates flexible amino acid side chains
        in `receptor`.
        Each residue, where at least one atom index is ``True`` in
        `flexible`, is considered flexible.
        By default, the receptor has no flexibility.
    bin_path : str, optional
        Path to the *Vina* binary.

    Examples
    --------

    >>> # A dummy receptor and ligand
    >>> ligand = residue("ASP")
    >>> receptor = atom_array
    >>> app = VinaApp(
    ...     ligand, receptor,
    ...     # Binding pocket is in the center of the receptor
    ...     center=centroid(receptor),
    ...     # 20 Å x 20 Å x 20 Å search space
    ...     size=[20, 20, 20],
    ...     # Handle residues 2 and 5 as flexible
    ...     flexible=(receptor.res_id == 2) | (receptor.res_id == 5)
    ... )
    """
    def __init__(self, ligand, receptor, center, size, flexible=None,
                 bin_path="vina"):
        super().__init__(bin_path)

        if ligand.bonds is None:
            raise ValueError("The ligand has no associated BondList")
        if receptor.bonds is None:
            raise ValueError("The receptor has no associated BondList")

        self._ligand = ligand.copy()
        self._receptor = receptor.copy()
        self._center = copy.deepcopy(center)
        self._size = copy.deepcopy(size)
        self._is_flexible = flexible is not None
        self._seed = None
        self._exhaustiveness = None
        self._number = None
        self._energy_range = None

        if self._is_flexible:
            flexible_indices = np.where(flexible)[0]
            self._flex_res_starts = np.unique(get_residue_starts_for(
                receptor, flexible_indices
            ))
        
        self._ligand_file  = NamedTemporaryFile(
            "w", suffix=".pdbqt", delete=False
        )
        self._receptor_file  = NamedTemporaryFile(
            "w", suffix=".pdbqt", delete=False
        )
        self._receptor_flex_file  = NamedTemporaryFile(
            "w", suffix=".pdbqt", delete=False
        )
        self._out_file  = NamedTemporaryFile(
            "r", suffix=".pdbqt", delete=False
        )
    
    @requires_state(AppState.CREATED)
    def set_seed(self, seed):
        """
        Fix the seed for the random number generator to get
        reproducible results.

        By default, the seed is chosen randomly.

        Parameters
        ----------
        seed : int
            The seed for the random number generator.
        """
        self._seed = seed
    
    @requires_state(AppState.CREATED)
    def set_exhaustiveness(self, exhaustiveness):
        """
        Set the *exhaustiveness* parameter for *Vina*.

        A higher exhaustiveness may lead to better docking results, but
        also increases the computation time.
        By default, the exhaustiveness is ``8``.

        Parameters
        ----------
        exhaustiveness : int
            The value for the exhaustiveness parameter.
            Must be greater than 0.
        """
        self._exhaustiveness = exhaustiveness
    
    @requires_state(AppState.CREATED)
    def set_max_number_of_models(self, number):
        """
        Set the maximum number of binding modes to generate.

        *Vina* may generate less modes, if the docking process does
        not find enough distinct conformations.
        By default, the maximum number is ``9``.

        Parameters
        ----------
        number : int
            The maximum number of generated modes/models.
        """
        self._number = number
    
    @requires_state(AppState.CREATED)
    def set_energy_range(self, energy_range):
        """
        Set the maximum energy range of the generated models.

        *Vina* will ignore binding modes if the difference between this
        mode and the best mode is greater than this value.
        By default, the range is ``3.0``.

        Parameters
        ----------
        number : float
            The energy range (kcal/mol).
        """
        self._energy_range = energy_range

    def run(self):
        # Use different atom ID ranges for atoms in ligand and receptor
        # for unambiguous assignment, if the receptor contains flexible
        # residues
        self._ligand.set_annotation("atom_id", np.arange(
            1,
            self._ligand.array_length() + 1
        ))
        self._receptor.set_annotation("atom_id", np.arange(
            self._ligand.array_length() + 1,
            self._ligand.array_length() + self._receptor.array_length() + 1
        ))

        ligand_file = PDBQTFile()
        # Contains 'true' entries for all atoms that have not been 
        # removed from ligand
        self._ligand_mask = ligand_file.set_structure(
            self._ligand,
            rotatable_bonds="all"
        )
        ligand_file.write(self._ligand_file)
        self._ligand_file.flush()
        
        if self._is_flexible:
            self._rigid_mask = np.ones(
                self._receptor.array_length(), dtype=bool
            )
            # Contains 'true' entries for all atoms that have not been 
            # removed from receptor in flexible side chains
            self._receptor_mask = np.zeros(
                self._receptor.array_length(), dtype=bool
            )
            for i, start in enumerate(self._flex_res_starts):
                flex_mask, rigid_mask, root = self._get_flexible_residue(start)
                self._rigid_mask &= rigid_mask
                root_in_flex_residue = np.where(
                    np.arange(self._receptor.array_length())[flex_mask] == root
                )[0][0]
                flex_file = PDBQTFile()
                self._receptor_mask[flex_mask] |= flex_file.set_structure(
                    self._receptor[flex_mask],
                    rotatable_bonds="all",
                    root=root_in_flex_residue,
                    include_torsdof=False
                )
                # Enclose each flexible residue
                # with BEGIN_RES and END_RES
                self._receptor_flex_file.write(f"BEGIN_RES {i}\n")
                flex_file.write(self._receptor_flex_file)
                self._receptor_flex_file.write(f"END_RES {i}\n")
            self._receptor_flex_file.flush()

            receptor_file = PDBQTFile()
            receptor_file.set_structure(
                self._receptor[self._rigid_mask],
                rotatable_bonds=None,
                include_torsdof=False
            )
            receptor_file.write(self._receptor_file)
            self._receptor_file.flush()

        else:
            receptor_file = PDBQTFile()
            receptor_file.set_structure(
                self._receptor,
                rotatable_bonds=None,
                include_torsdof=False
            )
            receptor_file.write(self._receptor_file)
            self._receptor_file.flush()

        arguments = [
            "--ligand", self._ligand_file.name,
            "--receptor", self._receptor_file.name,
            "--out", self._out_file.name,
            "--center_x", f"{self._center[0]:.3f}",
            "--center_y", f"{self._center[1]:.3f}",
            "--center_z", f"{self._center[2]:.3f}",
            "--size_x", f"{self._size[0]:.3f}",
            "--size_y", f"{self._size[1]:.3f}",
            "--size_z", f"{self._size[2]:.3f}",
        ]
        if self._seed is not None:
            arguments.extend(["--seed", str(self._seed)])
        if self._exhaustiveness is not None:
            arguments.extend(["--exhaustiveness", str(self._exhaustiveness)])
        if self._number is not None:
            arguments.extend(["--num_modes", str(self._number)])
        if self._energy_range is not None:
            arguments.extend(["--energy_range", str(self._energy_range)])
        if self._is_flexible:
            arguments.extend(["--flex", str(self._receptor_flex_file.name)])

        self.set_arguments(arguments)
        super().run()
    
    def evaluate(self):
        super().evaluate()
        out_file = PDBQTFile.read(self._out_file)
        
        models = out_file.get_structure()

        n_ligand_atoms = np.count_nonzero(self._ligand_mask)
        self._ligand_models = models[..., :n_ligand_atoms]
        self._flex_models = models[..., n_ligand_atoms:]
        self._n_models = models.stack_depth()
        
        remarks = out_file.get_remarks()
        self._energies = np.array(
            # VINA RESULT:      -5.8      0.000      0.000
            #                     ^
            [float(remark[12:].split()[0]) for remark in remarks]
        )
    
    def clean_up(self):
        super().clean_up()
        cleanup_tempfile(self._ligand_file)
        cleanup_tempfile(self._receptor_file)
        cleanup_tempfile(self._receptor_flex_file)
        cleanup_tempfile(self._out_file)
    
    @requires_state(AppState.JOINED)
    def get_energies(self):
        """
        Get the predicted binding energy for each generated binding
        mode.

        Returns
        -------
        energies : ndarray, dtype=float
            The predicted binding energies (kcal/mol).
            The energies are sorted from best to worst.
        """
        return self._energies

    @requires_state(AppState.JOINED)
    def get_ligand_models(self):
        """
        Get the ligand structure with the conformations for each 
        generated binding mode.

        Returns
        -------
        ligand : AtomArrayStack
            The docked ligand.
            Each model corresponds to one binding mode.
            The models are sorted from best to worst predicted binding
            affinity.
        
        Notes
        -----
        The returned structure may contain less atoms than the input
        structure, as *Vina* removes nonpolar hydrogen atoms.
        Furthermore, the returned structure contains *AutoDock* atom
        types as ``element`` annotation.
        """
        return self._ligand_models

    @requires_state(AppState.JOINED)
    def get_ligand_coord(self):
        """
        Get the ligand coordinates for each generated binding mode.

        Returns
        -------
        coord : ndarray, shape=(m,n,3), dtype=float
            The coordinates for *m* binding modes and *n* atoms
            of the input ligand.
            The models are sorted from best to worst predicted binding
            affinity.
            Missing coordinates due to the removed nonpolar hydrogen
            atoms are set to *NaN*.
        """
        coord = np.full(
            (self._n_models, self._ligand.array_length(), 3),
            np.nan, dtype=np.float32
        )
        coord[:, self._ligand_mask] = self._ligand_models.coord
        return coord
    
    @requires_state(AppState.JOINED)
    def get_flexible_residue_models(self):
        """
        Get the structure for the flexible side chains with the
        conformations for each generated binding mode.

        If no flexible side chains were defined, the returned
        :class:`AtomArrayStack` contains no atoms.

        Returns
        -------
        side_chains : AtomArrayStack
            The docked side chains.
            Each model corresponds to one binding mode.
            The models are sorted from best to worst predicted binding
            affinity.
        
        Notes
        -----
        The returned structure may contain less atoms than the input
        structure, as *Vina* removes nonpolar hydrogen atoms.
        Furthermore, the returned structure contains *AutoDock* atom
        types as ``element`` annotation.
        """
        return self._flex_models

    @requires_state(AppState.JOINED)
    def get_receptor_coord(self):
        """
        Get the get_receptor_coord coordinates for each generated
        binding mode.

        Returns
        -------
        coord : ndarray, shape=(m,n,3), dtype=float
            The coordinates for *m* binding modes and *n* atoms
            of the input receptor.
            The models are sorted from best to worst predicted binding
            affinity.
            Missing coordinates due to the removed nonpolar hydrogen
            atoms from flexible side chains are set to *NaN*.
        
        Notes
        -----
        The output is only meaningful, if flexible side chains were
        defined.
        Otherwise, the returned coordinates are simply *m* repetitions
        of the input receptor coordinates.
        """
        coord = np.repeat(
            self._receptor.coord[np.newaxis, ...],
            repeats=self._n_models, axis=0
        )
        if self._is_flexible:
            # Replace original coordinates with modeled coordinates
            # for the the flexible side chains
            # The coordinates from removed atoms are NaN
            coord[:, ~self._rigid_mask] = np.nan
            coord[:, self._receptor_mask] = self._flex_models.coord
        return coord

    def _get_flexible_residue(self, residue_start):
        residue_indices = np.where(
            get_residue_masks(self._receptor, [residue_start])[0]
        )[0]
        root_indices_in_residue = np.isin(
            self._receptor.atom_name[residue_indices], ("CA",)
        )
        root_indices = residue_indices[root_indices_in_residue]
        if len(root_indices) == 0:
            raise BadStructureError("Found no CA atom in residue")
        if len(root_indices) > 1:
            raise BadStructureError("Multiple CA atom in residue")
        root_index = root_indices[0]

        # Find the index of the atom connected to root on the flexible
        # side chain (CB)
        root_connect_indices, _ = self._receptor.bonds.get_bonds(root_index)
        connected_index = None
        try:
            connected_index = root_connect_indices[np.isin(
                self._receptor.atom_name[root_connect_indices], ("CB",)
            )][0]
        except IndexError:
            # Residue has no appropriate connection (e.g. in glycine)
            # -> There is no atom in the flexible side chain
            flex_mask = np.zeros(self._receptor.array_length(), dtype=bool)
            rigid_mask = np.ones(self._receptor.array_length(), dtype=bool)
            return flex_mask, rigid_mask, root_index
        
        # Remove the root bond from the bond list
        # to find the atoms involved in the flexible part
        bonds = self._receptor.bonds.copy()
        bonds.remove_bond(root_index, connected_index)
        flexible_indices = find_connected(bonds, connected_index)
        if root_index in flexible_indices:
            raise BadStructureError(
                "There are multiple connections between the flexible and "
                "rigid part, maybe a cyclic residue like proline was selected" 
            )

        flex_mask = np.zeros(self._receptor.array_length(), dtype=bool)
        flex_mask[flexible_indices] = True
        rigid_mask = ~flex_mask
        # Root index is part of rigid and flexible part
        flex_mask[root_index] = True

        return flex_mask, rigid_mask, root_index
    

    @staticmethod
    def dock(ligand, receptor, center, size, flexible=None, bin_path="vina"):
        """
        Dock a ligand to a receptor molecule using *AutoDock Vina*.

        This is a convenience function, that wraps the :class:`VinaApp`
        execution.

        Parameters
        ----------
        ligand : AtomArray
            The structure of the receptor molecule.
            Must have an associated :class:`BondList`.
            An associated ``charges`` annotation is recommended for proper
            calculation of partial charges.
        receptor : AtomArray, shape=(n,)
            The structure of the receptor molecule.
            Must have an associated :class:`BondList`.
            An associated ``charges`` annotation is recommended for proper
            calculation of partial charges.
        center : ndarray, shape=(3,), dtype=float
            The *xyz* coordinates for the center of the search space.
        size : ndarray, shape=(3,), dtype=float
            The size of the search space in *xyz* directions.
        flexible : ndarray, shape=(n,), dtype=bool, optional
            A boolean mask that indicates flexible amino acid side chains
            in `receptor`.
            Each residue, where at least one atom index is ``True`` in
            `flexible`, is considered flexible.
            By default, the receptor has no flexibility.
        bin_path : str, optional
            Path to the *Vina* binary.

        Returns
        -------
        coord : ndarray, shape=(m,n,3), dtype=float
            The docked ligand coordinates for *m* binding modes and
            *n* atoms of the input ligand.
            The models are sorted from best to worst predicted binding
            affinity.
            Missing coordinates due to the removed nonpolar hydrogen
            atoms are set to *NaN*.
        energies : ndarray, shape=(m,), dtype=float
            The corresponding predicted binding energies (kcal/mol).
        """
        app = VinaApp(ligand, receptor, center, size, flexible, bin_path)
        app.start()
        app.join()
        return app.get_ligand_coord(), app.get_energies()



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.align"
__author__ = "Patrick Kunzmann"
__all__ = ["EValueEstimator"]

import numpy as np
from ..seqtypes import GeneralSequence
from .pairwise import align_optimal


class EValueEstimator:
    r"""
    This class is used to calculate *expect values* (E-values) for local
    pairwise sequence alignments.

    The E-value is a measure to quantify the significance of a found
    homology.
    It is the number of alignments, that would result from aligning
    random sequences of a given length, with a score at least as high as
    the score from an alignment of interest.

    The calculation of the E-value from score and sequence lengths
    depend on the two parameters :math:`\lambda` and :math:`K`
    :footcite:`Altschul1996`.
    These parameters are estimated from sampling a large number
    of random sequence alignments in :meth:`from_samples()`
    :footcite:`Altschul1986`, which may be time consuming.
    If these parameters are known, the constructor can be used instead.
    
    Based on the sampled parameters, the decadic logarithm of the
    E-value can be quickly calculated via :meth:`log_evalue()`.

    Parameters
    ----------
    lam : float
        The :math:`\lambda` parameter.
    k : float
        The :math:`K` parameter.
    
    Notes
    -----
    The calculated E-value is a rough estimation that gets more
    accurate the more sequences are used in the sampling process.
    Note that the accuracy for alignment of short sequences, where the
    average length of a sampled alignment make up a significant part of
    the complete sampled sequence :footcite:`Altschul1996`.

    References
    ----------

    .. footbibliography::

    Examples
    --------

    Create an alignment, whose significance should be evaluated.

    >>> query = NucleotideSequence("CGACGGCGTCTACGAGTCAACATCATTC")
    >>> hit = NucleotideSequence("GCTTTATTACGGGTTTACGAGTTCAACATCACGAAAACAA")
    >>> matrix = SubstitutionMatrix.std_nucleotide_matrix()
    >>> gap_penalty = (-12, -2)
    >>> alignment = align_optimal(query, hit, matrix, gap_penalty, local=True)[0]
    >>> print(alignment)
    ACGGCGTCTACGAGT-CAACATCA
    ACGG-GTTTACGAGTTCAACATCA
    >>> print(alignment.score)
    77

    Create an estimator based on the same scoring scheme as the
    alignment.
    Use background symbol frequencies from the hypothetical reference
    database.

    >>> # Ensure deterministic results
    >>> np.random.seed(0)
    >>> # Sequences in database have a GC content of 0.6
    >>> background = np.array([0.2, 0.3, 0.3, 0.2])
    >>> estimator = EValueEstimator.from_samples(
    ...     query.alphabet, matrix, gap_penalty, background, sample_length=100
    ... )

    Approach 1: Calculate E-value based on number of sequences in the
    hypothetical database (*100*).

    >>> log_e = estimator.log_evalue(alignment.score, len(query), 100 * len(hit))
    >>> print(f"E-value = {10**log_e:.2e}")
    E-value = 3.36e-01

    Approach 2: Calculate E-value based on total length of all sequences
    in the hypothetical database combined (*10000*).

    >>> log_e = estimator.log_evalue(alignment.score, len(query), 10000)
    >>> print(f"E-value = {10**log_e:.2e}")
    E-value = 8.41e-01
    """

    def __init__(self, lam, k):
        self._lam = lam
        self._k = k

    @staticmethod
    def from_samples(alphabet, matrix, gap_penalty, frequencies,
                     sample_length=1000, sample_size=1000):
        r"""
        Create an :class:`EValueEstimator` with :math:`\lambda` and
        :math:`K` estimated via sampling alignments of random sequences
        based on a given scoring scheme.

        The parameters are estimated from the sampled alignment scores
        using the method of moments :footcite:`Altschul1986`.

        Parameters
        ----------
        alphabet : Alphabet, length=k
            The alphabet for the sampled sequences.
        matrix : SubstitutionMatrix
            The substitution matrix.
            It must be compatible with the given `alphabet` and the
            expected similarity score between two random symbols must be
            negative.
        gap_penalty : int or tuple(int,int)
            Either a linear (``int``) or affine (``tuple``) gap penalty.
            Integers must be negative.
        frequencies : ndarray, shape=k, dtype=float
            The background frequencies for each symbol in the
            `alphabet`.
            The random sequences are created based on these frequencies.
        sample_length : int
            The length of the sampled sequences.
            It should be much larger than the average length of a local
            alignment of two sequences.
            The runtime scales quadratically with this parameter.
        sample_size : int
            The number of sampled sequences.
            The accuracy of the estimated parameters and E-values,
            but also the runtime increases with the sample size.
        
        Returns
        -------
        estimator : EValueEstimator
            A :class:`EValueEstimator` with sampled :math:`\lambda` and
            :math:`K` parameters.
        
        Notes
        -----
        The sampling process generates random sequences based on
        ``numpy.random``.
        To ensure reproducible results you could call
        :func:`numpy.random.seed()` before running
        :meth:`from_samples()`.
        """
        if len(frequencies) != len(alphabet):
            raise IndexError(
                f"Background frequencies for {len(frequencies)} symbols were "
                f"given, but the alphabet has {len(alphabet)} symbols"
            )
        if np.any(frequencies < 0):
            raise ValueError("Background frequencies must be positive")
        # Normalize background frequencies
        frequencies = frequencies / np.sum(frequencies)

        # Check matrix
        if not matrix.is_symmetric():
            raise ValueError("A symmetric substitution matrix is required")
        if not matrix.get_alphabet1().extends(alphabet):
            raise ValueError(
                "The substitution matrix is not compatible "
                "with the given alphabet"
            )
        score_matrix = matrix.score_matrix()[:len(alphabet), :len(alphabet)]
        if np.sum(
            score_matrix \
            * frequencies[np.newaxis, :] \
            * frequencies[:, np.newaxis]
        ) >= 0:
            raise ValueError(
                "Invalid substitution matrix, the expected similarity "
                "score between two random symbols is not negative"
            )

        # Generate the sequence code for the random sequences
        random_sequence_code = np.random.choice(
            len(alphabet),
            size=(sample_size, 2, sample_length),
            p=frequencies
        )

        # Sample the alignments of random sequences
        sample_scores = np.zeros(sample_size, dtype=int)
        for i in range(sample_size):
            seq1 = GeneralSequence(alphabet)
            seq2 = GeneralSequence(alphabet)
            seq1.code = random_sequence_code[i,0]
            seq2.code = random_sequence_code[i,1]
            sample_scores[i] = align_optimal(
                seq1, seq2, matrix,
                local=True, gap_penalty=gap_penalty, max_number=1
            )[0].score
        
        # Use method of moments to estimate parameters
        lam = np.pi / np.sqrt(6 * np.var(sample_scores))
        u = np.mean(sample_scores) - np.euler_gamma / lam
        k = np.exp(lam * u) / sample_length**2
        
        return EValueEstimator(lam, k)

    @property
    def lam(self):
        return self._lam
    
    @property
    def k(self):
        return self._k
    
    def log_evalue(self, score, seq1_length, seq2_length):
        r"""
        Calculate the decadic logarithm of the E-value for a given
        score.

        The E-value and the logarithm of the E-value is calculated as

        .. math::
        
            E = Kmn e^{-\lambda s}

            \log_{10} E = (\log_{10} Kmn) - \frac{\lambda s}{\ln 10},
        
        where :math:`s` is the similarity score and :math:`m` and
        :math:`n` are the lengths of the aligned sequences.

        Parameters
        ----------
        score : int or ndarray, dtype=int
            The score to evaluate.
        seq1_length : int or ndarray, dtype=int
            The length of the first sequence.
            In the context of a homology search in a sequence database,
            this is usually the length of the query sequence.
        seq2_length : int or ndarray, dtype=int
            The length of the second sequence.
            In the context of a homology search in a sequence database,
            this is usually either the combined length of all sequences
            in the database or the length of the hit sequence multiplied
            by the number of sequences in the database.
        
        Returns
        -------
        log_e : float
            The decadic logarithm of the E-value.
        
        Notes
        -----
        This method returns the logarithm of the E-value instead of
        the E-value, as low E-values indicating a highly significant
        homology cannot be accurately represented by a ``float``.
        """
        score = np.asarray(score)
        seq1_length = np.asarray(seq1_length)
        seq2_length = np.asarray(seq2_length)

        return np.log10(self._k * seq1_length * seq2_length) \
            - self._lam * score / np.log(10)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.graphics"
__author__ = "Patrick Kunzmann"
__all__ = ["get_color_scheme", "list_color_scheme_names", "load_color_scheme"]

import numpy as np
import json
from os.path import join, dirname, realpath
import glob
import os
from ..alphabet import Alphabet


def load_color_scheme(file_name):
    """
    Load a color scheme from a JSON file.

    A color scheme is a list of colors that correspond to symbols of an
    alphabet. The color for a symbol is list of colors indexed by the
    corresponding symbol code.

    Parameters
    ----------
    file_name : str
        The file name of the JSON file containing the scheme.
    
    Returns
    -------
    scheme : dict
        A dictionary representing the color scheme, It contains the
        following keys, if the input file is proper:
        
           - **name** - Name of the scheme.
           - **alphabet** - :class:`Alphabet` instance describing the
             type of sequence the scheme can be used for.
           - **colors** - List of *Matplotlib* compatible colors
    """
    with open(file_name, "r") as file:
        scheme = json.load(file)
        alphabet = Alphabet(scheme["alphabet"])
        # Store alphabet as 'Alphabet' object
        scheme["alphabet"] = alphabet
        colors = [None] * len(alphabet)
        for key, value in scheme["colors"].items():
            index = alphabet.encode(key)
            colors[index] = value
        # Store colors as symbol code ordered list of colors,
        # rather than dictionary
        scheme["colors"] = colors
        return scheme


def get_color_scheme(name, alphabet, default="#FFFFFF"):
    """
    Get a color scheme by name and alphabet.

    A color scheme is a list of colors that correspond to symbols of an
    alphabet. The color for a symbol is list of colors indexed by the
    corresponding symbol code.

    Parameters
    ----------
    name : str
        The name of the color scheme.
    alphabet : Alphabet
        The alphabet to obtain the scheme for. The alphabet of the
        scheme must equal or extend this parameter.
    default : str or tuple, optional
        A *Matplotlib* compatible color that is used for symbols that
        have no defined color in the scheme.
    
    Returns
    -------
    colors : list
        A list of *Matplotlib* compatible colors. The colors in the list
        have the same order as the symbols in the given `alphabet`.
        Since the alphabet of the color scheme may extend the given
        `alphabet`, the list of colors can be longer than the
        `alphabet`.

    Notes
    -----
    There can be multiple color schemes with the same name but for
    different alphabets (e.g. one for dna and one for protein
    sequences).

    Examples
    --------

    >>> alphabet = NucleotideSequence.alphabet_unamb
    >>> color_scheme = get_color_scheme("rainbow", alphabet)
    >>> print(color_scheme)
    ['#3737f5', '#37f537', '#f5f537', '#f53737']
    """
    for scheme in _color_schemes:
        if scheme["name"] == name and scheme["alphabet"].extends(alphabet):
            colors = scheme["colors"]
            # Replace None values with default color
            colors = [color if color is not None else default
                      for color in colors]
            # Only return colors that are in scope of this alphabet
            # and not the extended alphabet
            return colors[:len(alphabet)]
    raise ValueError(f"Unkown scheme '{name}' for given alphabet")


def list_color_scheme_names(alphabet):
    """
    Get a list of available color scheme names for a given alphabet.

    Parameters
    ----------
    alphabet : Alphabet
        The alphbet to get the color scheme names for.
        The alphabet of the scheme must equal or extend this parameter,
        to be included in the list.
    
    Returns
    -------
    schemes : list of str
        A list of available color schemes.
    """
    scheme_list = []
    for scheme in _color_schemes:
        if scheme["alphabet"].extends(alphabet):
            scheme_list.append(scheme["name"])
    return scheme_list


_scheme_dir = join(dirname(realpath(__file__)), "color_schemes")

_color_schemes = []

for file_name in glob.glob(_scheme_dir + os.sep + "*.json"):
    scheme = load_color_scheme(file_name)
    _color_schemes.append(scheme)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.graphics"
__author__ = "Patrick Kunzmann"
__all__ = ["plot_atoms", "plot_ball_and_stick_model"]

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Line3DCollection


def plot_atoms(axes, atoms, colors, line_width=1.0, background_color=None,
               center=None, size=None, zoom=1.0):
    """
    Plot an :class:`AtomArray` as lines between bonded atoms.

    The z-axis points into the screen plane.
    
    Parameters
    ----------
    axes : Axes3D
        The *Matplotlib* 3D-axes to plot the structure on.
    atoms : AtomArray, shape=(n,)
        The structure to be plotted.
        The atom array must have an associated :class:`BondList`, i.e.
        its ``bonds`` attribute must be defined.
        One line for each bond is drawn.
    colors : ndarray, shape=(n,3) or shape=(n,4), dtype=float
        An array of RGB or RGBA colors for each atom in `atoms`.
        The values for each color channel are in the range 0 to 1.
    line_width : float, optional
        The width of the lines to be drawn.
    background_color : object
        A *Matplotlib* compatible color (color name or RGB values).
        If set, the background is colored with the given value.
    center : tuple of (float, float, float), optional
        The coordinates of the structure that are in the center of the
        plot.
        By default the complete molecule is centered.
    size : float, optional
        The size of each dimension in the plot.
        The limits of the :class:`Axes3D` are set to
        ``(center - size/(2*zoom)), (center + size/(2*zoom))``.
    zoom : float, optional
        Zoom in or zoom out.

        - ``> 1.0``: Zoom in.
        - ``< 1.0``: Zoom out.
    
    Notes
    -----
    This is a very simple visualization tools for quick visual analysis
    of a structure.
    For publication-ready molecular images the usage of a dedicated
    molecular visualization tool is recommended.
    """
    if not isinstance(axes, Axes3D):
        raise ValueError("The given axes mut be an 'Axes3D'")
    if atoms.bonds is None:
        raise ValueError("The atom array must have an associated bond list")
    
    # Calculating connections between atoms
    line_coord = []
    line_colors = []
    for index1, index2 in atoms.bonds.as_array()[:,:2]:
        # Every connection consist of two lines:
        # One from the first atom to the center
        # and from from the second atom to the center
        line_start = atoms.coord[index1]
        line_end = atoms.coord[index2]
        line_center = (line_start + line_end) / 2
        
        # Add line from first atom
        line_coord.append((
            line_start, line_center
        ))
        line_colors.append(colors[index1])
        
        # Add line from second atom
        line_coord.append((
            line_end, line_center
        ))
        line_colors.append(colors[index2])

    # Plot computed line coordinates and colors
    # Use 'Line3DCollection' for higher efficiency
    lines = Line3DCollection(
        line_coord, color=line_colors, linewidths=line_width
    )
    axes.add_collection(lines)
    
    # Set viewing angle
    axes.azim = -90
    axes.elev = 90
    # Remove frame
    axes.axis("off")
    # Set background color
    if background_color is not None:
        axes.set_facecolor(background_color)
        axes.get_figure().set_facecolor(background_color)
    _set_box(axes, atoms.coord, center, size, zoom)


def plot_ball_and_stick_model(axes, atoms, colors, ball_size=200,
                              line_color="black", line_width=1.0,
                              background_color=None, center=None,
                              size=None, zoom=1.0):
    """
    Plot an :class:`AtomArray` as *ball-and-stick* model.

    The z-axis points into the screen plane.

    UNSTABLE: This function is probably subject to future changes.
    
    Parameters
    ----------
    axes : Axes3D
        The *Matplotlib* 3D-axes to plot the structure on.
    atoms : AtomArray, length=n
        The structure to be plotted.
        The atom array must have an associated :class:`BondList`, i.e.
        its ``bonds`` attribute must be defined.
        One line for each bond is drawn.
    colors : ndarray, shape=(n,3) or shape=(n,4), dtype=float
        An array of RGB or RGBA colors for each atom in `atoms`, that
        is used to color the *balls* of the model.
        The values for each color channel are in the range 0 to 1.
    ball_size : int or iterable of int, shape=(n,)
        The size of the *balls* in the model in *pt*.
        Either a single value for all atoms or an iterable object of
        values for each atom.
    line_color : object
        A *Matplotlib* compatible color value for the *sticks* of the
        model.
    line_width : float, optional
        The width of the *sticks* in the model in *pt*.
    background_color : object
        A *Matplotlib* compatible color (color name or RGB values).
        If set, the background is colored with the given value.
    center : tuple of (float, float, float), optional
        The coordinates of the structure that are in the center of the
        plot.
        By default the complete molecule is centered.
    size : float, optional
        The size of each dimension in the plot.
        The limits of the :class:`Axes3D` are set to
        ``(center - size/(2*zoom)), (center + size/(2*zoom))``.
    zoom : float, optional
        Zoom in or zoom out.

        - ``> 1.0``: Zoom in.
        - ``< 1.0``: Zoom out.
    
    Notes
    -----
    This is a very simple visualization tools for quick visual analysis
    of a structure.
    For publication-ready molecular images the usage of a dedicated
    molecular visualization tool is recommended.
    """
    if not isinstance(axes, Axes3D):
        raise ValueError("The given axes mut be an 'Axes3D'")
    if atoms.bonds is None:
        raise ValueError("The atom array must have an associated bond list")
    
    # Calculating connections between atoms
    line_coord = [
        (atoms.coord[index1], atoms.coord[index2])
        for index1, index2 in atoms.bonds.as_array()[:,:2]
    ]

    # Plot sticks
    # Use 'Line3DCollection' for higher efficiency
    sticks = Line3DCollection(
        line_coord, color=line_color, linewidths=line_width
    )
    axes.add_collection(sticks)

    # Plot balls
    axes.scatter(
        *atoms.coord.T, s=ball_size, c=colors, linewidth=0, alpha=1
    )
    
    # Set viewing angle
    axes.azim = -90
    axes.elev = 90
    # Remove frame
    axes.axis("off")
    # Set background color
    if background_color is not None:
        axes.set_facecolor(background_color)
        axes.get_figure().set_facecolor(background_color)
    _set_box(axes, atoms.coord, center, size, zoom)


def _set_box(axes, coord, center, size, zoom):
    """
    This ensures an approximately equal aspect ratio in a 3D plot under
    the condition, that the :class:`Axes` is quadratic on the display.
    """
    if center is None:
        center = (
            (coord[:, 0].max() + coord[:, 0].min()) / 2,
            (coord[:, 1].max() + coord[:, 1].min()) / 2,
            (coord[:, 2].max() + coord[:, 2].min()) / 2,
        )

    if size is None:
        size = np.array([
            coord[:, 0].max() - coord[:, 0].min(),
            coord[:, 1].max() - coord[:, 1].min(),
            coord[:, 2].max() - coord[:, 2].min()
        ]).max()
    
    axes.set_xlim(center[0] - size/(2*zoom), center[0] + size/(2*zoom))
    axes.set_ylim(center[1] - size/(2*zoom), center[1] + size/(2*zoom))
    axes.set_zlim(center[2] - size/(2*zoom), center[2] + size/(2*zoom))
    
    # Make the axis lengths of the 'plot box' equal
    # The 'plot box' is not visible due to 'axes.axis("off")'
    axes.set_box_aspect([1,1,1])


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application"
__author__ = "Patrick Kunzmann"
__all__ = ["WebApp", "RuleViolationError"]

import abc
from .application import Application


class WebApp(Application, metaclass=abc.ABCMeta):
    """
    The base class for all web based applications.
    
    It allows for getting and setting the URL of the app and raises
    an :class:`RuleViolationError` when a subclass calls
    :func:`violate_rule()`
    (e.g. when the server was contacted too often.)
    
    Be careful, when calling func:`get_app_state()`. This may involve a
    server contact and therefore frequent calls may raise a
    :class:`RuleViolationError`.
    
    Parameters
    ----------
    app_url : str
        URL of the web app.
    obey_rules : bool, optional
        If true, the application raises an :class:`RuleViolationError`, if
        the server rules are violated. (Default: True)
    """
    
    def __init__(self, app_url, obey_rules=True):
        super().__init__()
        self._obey_rules = obey_rules
        self._app_url = app_url
    
    def violate_rule(self, msg=None):
        """
        Indicate that a server rule was violated, i.e. this raises a
        :class:`RuleViolationError` unless `obey_rules` is false.
        
        PROTECTED: Do not call from outside.
        
        Parameters
        ----------
        msg : str, optional
            A custom message for the :class:`RuleViolationError`.
        """
        if self._obey_rules:
            if msg is None:
                raise RuleViolationError(
                    "The user guidelines would be violated"
                )
            else:
                raise RuleViolationError(msg)
    
    def app_url(self):
        """
        Get the URL of the web app.
        
        Returns
        -------
        url : str
            URL of the web app.
        """
        return self._app_url


class RuleViolationError(Exception):
    """
    Indicates that the user guidelines of the web application would be
    violated, if the program continued.
    """
    pass


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.graphics"
__author__ = "Patrick Kunzmann"
__all__ = ["plot_plasmid_map"]

import copy
import warnings
import abc
import numpy as np
import re
from ...visualize import colors
from ..annotation import Annotation, Feature, Location


def plot_plasmid_map(axes, annotation, plasmid_size, tick_length=0.02,
                     tick_step=200, ring_width=0.01, feature_width=0.06,
                     spacing=0.01, arrow_head_length=0.04, label=None,
                     face_properties=None, label_properties=None,
                     omit_oversized_labels=True, feature_formatter=None):
    """
    Plot a plasmid map using the sequence features in the given
    :class:`Annotation`.

    Each feature location is depicted either by a curved arrow,
    indicating on which strand the feature resides, or an arc.
    The ticks indicate the sequence position.

    **EXPERIMENTAL**: The API and the resulting plots will probably
    change in future versions.

    Parameters
    ----------
    axes : PolarAxes
        A *Matplotlib* axes, that is used as plotting area.
        A polar projection is required.
    annotation : Annotation
        The annotation to be visualized.
    plasmid_size : int
        The size of the plasmid, i.e. the length of its sequence.
    tick_length : float, optional
        The length of the axis ticks as percentage of the plot radius.
    tick_step : int, optional
        The interval between the displayed sequence positions.
        For example ``tick_step=200`` means that ticks will be displayed
        at ``1``, ``200``, ``400``, ``600``, etc.
    ring_width : float, optional
        The width of the outer ring as percentage of the plot radius.
    feature_width : float, optional
        The width of each feature arrow/arc as percentage of the
        plot radius.
    spacing : float, optional
        The spacing between the rows of feature arrows/arcs as
        percentage of the plot radius.
    arrow_head_length : float, optional
        The length of the arrow heads as percentage of the radius
        at which the respective arrow is plotted.
    label : str, optional
        The central label of the plot. Ususally the plasmid name.
    face_properties : dict, optional
        A dictionary of properties given to the patches that make up the
        feature arrows/arcs.
        Internally the arrow tail is a *Matplotlib* :class:`Rectangle`,
        and the arrow head is a :class:`Polygon`.
        For example this parameter could be used to add edge lines to
        the arrows/arcs.
    label_properties : dict, optional
        A dictionary of properties given to the feature labels.
        Internally each feature label is one or multiple *Matplotlib*
        :class:`Text` instances.
        For example this parameter could be used to set the font weight
        or size.
    omit_oversized_labels : bool, optional
        If true, a feature label will not be displayed, if the label is
        larger than its containing arc/arrow.
        This ensures, that labels from different features will not
        overlap each other.
        If false, the feature labels will always be displayed.
    feature_formatter : function, optional
        A function that determines *how* each feature is displayed.
        The given function must take a :class:`Feature` and must return
        the following tuple:

        - *directional* : bool
           
            True, if the direction of the feature should be indicated by
            an arrow.
            Otherwise, the feature is plotted is arc.
        
        - *face_color* : tuple or str, optional
            
            A *Matplotlib* compatible color for the feature arrow/arc.
        
        - *label_color* : tuple or str, optional
            
            A *Matplotlib* compatible color for the feature label.
        
        - *label* : str or None
            
            The label to be displayed for this feature.
            None, if no label should be displayed.
    """
    from matplotlib.projections.polar import PolarAxes
    
    if not isinstance(axes, PolarAxes):
        raise TypeError("The given axes must be a 'PolarAxes'")

    ### Setup parameters ###
    if plasmid_size is None:
        # 'stop' of 'get_location_range()' is exclusive -> -1
        plasmid_size = annotation.get_location_range()[1] - 1
    if face_properties is None:
        face_properties = {}
    if label_properties is None:
        label_properties = {}
    if feature_formatter is None:
        feature_formatter = _default_feature_formatter


    ### Setup matplotlib ###
    # The x-coordinate is given as angle (rad)
    # Full circle -> 2*pi
    axes.set_xlim(0, 2*np.pi)
    axes.set_ylim(0, 1)
    axes.yaxis.set_visible(False)
    axes.xaxis.set_tick_params(
        bottom=False, labelbottom=True
    )
    axes.set_theta_zero_location("N")
    axes.set_theta_direction("clockwise")
    axes.spines["polar"].set_visible(False)
    axes.grid(False)
    # Setup ticks
    ticks = [1]
    tick_labels = [str(1)]
    for tick in range(tick_step, plasmid_size, tick_step):
        ticks.append(tick)
        tick_labels.append(str(tick))
    # Sequence location is replaced by angle
    axes.xaxis.set_ticks([_loc_to_rad(tick, plasmid_size) for tick in ticks])
    axes.xaxis.set_ticklabels(tick_labels)
    ### Draw plasmid ring with ticks and central label ###
    
    # Plasmid ring
    # Use 'barh()' instead of a Rectangle patch to ensure that the axes
    # is properly initialized
    # Otherwise the feature rectangles are not curved, but straight
    axes.barh(
        1-ring_width-tick_length, 2*np.pi, ring_width,
        align="edge", color="black"
    )
    
    # Ticks (ticks itself, not the tick labels)
    for tick in ticks:
        angle = _loc_to_rad(tick, plasmid_size)
        axes.plot(
            (angle, angle), (1-tick_length, 1),
            color="black", linewidth=1, linestyle="-"
        )
    
    # Central plasmid label
    if label is not None:
        axes.text(
            0, 0, label, ha="center", va="center",
            color="black", size=32, fontweight="bold"
        )


    ### Draw plasmid interior ###
    inner_radius = 1 - ring_width - tick_length
    features = sorted(
        [
            _merge_over_periodic_boundary(feature, plasmid_size)
            for feature in annotation
        ],
        # Features are sorted by the length of their location range
        # The shortest come first
        key = lambda feature: np.diff(feature.get_location_range())[0],
        reverse = True
    )
    axes.add_artist(PlasmidMap(
        axes, 0, features, plasmid_size, inner_radius, feature_width, spacing,
        arrow_head_length, label, face_properties, label_properties,
        omit_oversized_labels, feature_formatter
    ))


try:
    # Only create these classes when matplotlib is installed
    from matplotlib.artist import Artist
    from matplotlib.transforms import Bbox
    from matplotlib.patches import Rectangle, Polygon


    class PlasmidMap(Artist):
        def __init__(self, axes, zorder, features, plasmid_size, radius,
                     feature_width, spacing, arrow_head_length, label,
                     face_properties, label_properties, omit_oversized_labels,
                     feature_formatter):
            super().__init__()
            self._axes = axes
            self.zorder = zorder
            self._features = features
            self._plasmid_size = plasmid_size
            self._radius = radius
            self._feature_width = feature_width
            self._spacing = spacing

            self._all_indicators = []
            for feature in features:
                indicators_for_feature = []
                for loc in feature.locs:
                     # Set proper positions in 'draw()' method
                    bbox = Bbox.from_extents(0, 0, 0, 0)
                    # Draw features as curved arrows (feature indicator)
                    indicator = axes.add_artist(Feature_Indicator(
                        axes, self.zorder + 1, feature, loc, bbox,
                        arrow_head_length, face_properties, label_properties,
                        omit_oversized_labels, feature_formatter
                    ))
                    indicators_for_feature.append(indicator)
                self._all_indicators.append(indicators_for_feature)

        
        def draw(self, renderer, *args, **kwargs):
            # Find the maximum amount of feature rows
            # (used for overlapping features)
            row_count = int(
                self._radius // (self._feature_width + self._spacing)
            )
            # Tracks the location ranges of feature that were added to
            # a row in order to check if that row is occupied
            ranges_in_row = [[] for i in range(row_count)]
            # Stores the bottom coordinate (radius) for each row
            row_bottoms = [
                self._radius - (row+1) * (self._feature_width + self._spacing)
                for row in range(row_count)
            ]

            # Arrange the feature indicators in an way,
            # that there is no overlap between them
            for feature, indicators_for_feature in zip(
                self._features, self._all_indicators
            ):
                row_bottom = None
                first, last = feature.get_location_range()

                for row_i, curr_range in enumerate(ranges_in_row):
                    is_occupied = False
                    if curr_range is not None:
                        # Check if row is occupied
                        for curr_first, curr_last in curr_range:
                            # If the location extends over periodic
                            # boundary the 'first' location is negative
                            if first > 0:
                                # 'Normal feature'
                                if first <= curr_last and last >= curr_first:
                                    is_occupied = True
                            else: # first < 1
                                # Location is over periodic boundary
                                if first + self._plasmid_size <= curr_last \
                                   or last >= curr_first:
                                        is_occupied = True
                    if not is_occupied:
                        # Row is not occupied by another feature
                        # in the location range of the new feature
                        # -> Use this row
                        if first > 0:
                            # 'Normal feature'
                            ranges_in_row[row_i].append((first, last))
                        else:
                            # Location is over periodic boundary
                            # Split into 'end' and 'start' part
                            ranges_in_row[row_i].append((
                                first + self._plasmid_size, self._plasmid_size
                            ))
                            ranges_in_row[row_i].append((
                                1, last
                            ))
                        row_bottom = row_bottoms[row_i]
                        break
                if row_bottom is None:
                    # No free row -> ignore feature and raise warning
                    warnings.warn(
                        "Too many feature overlaps, try to increase the "
                        "radius or decrease the feature width or spacing"
                    )
                else:
                    for loc, indicator in zip(
                        feature.locs, indicators_for_feature
                    ):
                        # Calculate arrow shape parameters
                        row_center = row_bottom + self._feature_width/2
                        row_top = row_bottom + self._feature_width
                        start_ang = _loc_to_rad(loc.first, self._plasmid_size)
                        stop_ang  = _loc_to_rad(loc.last, self._plasmid_size)
                        bbox = Bbox.from_extents(
                            start_ang, row_bottom, stop_ang, row_top
                        )
                        indicator.set_bbox(bbox)


    class Feature_Indicator(Artist):
        def __init__(self, axes, zorder, feature, loc, bbox, head_length,
                     arrow_properties, label_properties, omit_oversized_labels,
                     feature_formatter):
            super().__init__()
            self._axes = axes
            self.zorder = zorder
            self._direction = loc.strand
            self._bbox = bbox
            self._head_length = head_length
            self._omit_oversized_labels = omit_oversized_labels
            
            # Determine how to draw the feature
            directional, face_color, label_color, label \
                = feature_formatter(feature)
            
            # Draw arrow as composition of a rectangle and a triangle,
            # as FancyArrow does not properly work for polar plots

            self._arrow_tail = axes.add_patch(Rectangle(
                # Set positions in 'draw()' method
                (0, 0), 0, 0,
                # Line width is set to 1 to avoid strange artifact in
                # the transition from rectangle (tail) to polygon (head)
                color=face_color, linewidth=1, zorder = self.zorder + 1,
                **arrow_properties
            ))
            
            if directional:
                # Only draw any arrow head when feature has a direction,
                # otherwise simply draw the tail (rectangle)
                self._arrow_head = axes.add_patch(Polygon(
                    # Set positions in 'draw()' method
                    [(0, 0), (0, 0), (0, 0)],
                    color=face_color, linewidth=1, zorder = self.zorder + 1,
                    **arrow_properties
                ))
            else:
                self._arrow_head = None

            if label is not None:
                label_properties["color"] = label_color
                self._label = axes.add_artist(CurvedText(
                    # Set positions in 'draw()' method
                    axes, self.zorder + 1, 0, 0, label, label_properties
                ))
            else:
                self._label = None
        

        def set_bbox(self, bbox):
            self._bbox = bbox
            center_x = (bbox.x0 + bbox.x1) / 2
            center_y = (bbox.y0 + bbox.y1) / 2
            if self._label is not None:
                self._label.set_position(center_x, center_y)


        def draw(self, renderer, *args, **kwargs):
            bbox = self._bbox
            center_x = (bbox.x0 + bbox.x1) / 2
            center_y = (bbox.y0 + bbox.y1) / 2

            # Constant absolute width for all arrows
            # irrespective of the radius in the polar plot
            # Calculate actual angle from given absolute width
            head_length = self._head_length / center_y
            
            # Check if the head should be drawn
            if self._arrow_head is None:
                head_length = 0
            # Check if the feature location is too small for
            elif head_length > bbox.width:
                # Limit size of arrow head to range of location
                head_length = bbox.width

            if self._direction == Location.Strand.FORWARD:
                rect_pos = (bbox.x0, bbox.y0)
                # (x0, y0), (x1, y1), (x2, y2)
                triangle_coord = [
                    (bbox.x1 - head_length, bbox.y0), # base 1
                    (bbox.x1 - head_length, bbox.y1), # base 2
                    (bbox.x1,              center_y) # tip
                ]
            else:
                rect_pos = (bbox.x0+head_length, bbox.y0)
                triangle_coord = [
                    (bbox.x0 + head_length, bbox.y0), # base 1
                    (bbox.x0 + head_length, bbox.y1), # base 2
                    (bbox.x0,              center_y) # tip
                ]
            
            # Update coordinates of sub-artists
            self._arrow_tail.set_xy(rect_pos)
            self._arrow_tail.set_width(bbox.width-head_length)
            self._arrow_tail.set_height(bbox.height)
            if self._arrow_head is not None:
                self._arrow_head.set_xy(triangle_coord)
            
            if self._label is not None:
                # Do not draw the labels if it is larger than the
                # indicator
                if self._omit_oversized_labels \
                   and self._label.get_total_angle(renderer) > bbox.width:
                        self._label.set_visible(False)
                else:
                    self._label.set_visible(True)



    class CurvedText(Artist):
        def __init__(self, axes, zorder, angle, radius, string,
                     text_properties):
            super().__init__()
            self._axes = axes
            self.zorder = zorder
            self._angle = angle
            self._radius = radius

            self._texts = []
            for word in _split_into_words(string):
                text = axes.text(
                    # Set position in 'draw()' method
                    0, 0,
                    word,
                    ha="center", va="center",
                    zorder=self.zorder + 1,
                    **text_properties,
                )
                self._texts.append(text)
        

        def set_visible(self, visible):
            super().set_visible(visible)
            for text in self._texts:
                text.set_visible(visible)


        def set_position(self, angle, radius):
            self._angle = angle
            self._radius = radius
        

        def get_total_angle(self, renderer):
            return np.sum(self.get_word_angles(renderer))
        

        def get_word_angles(self, renderer):
            ax_px_radius = self._axes.get_window_extent(renderer).width / 2
            ax_unit_radius = self._axes.get_ylim()[1]
            circle_px_circumference = ax_px_radius * 2*np.pi \
                                      * (self._radius / ax_unit_radius)

            rad_angle = 360 - np.rad2deg(self._angle)
            # Avoid to draw the text upside down, when drawn on the
            # bottom half of the map
            if rad_angle > 90 and rad_angle < 270:
                turn_around = True
            else:
                turn_around = False
            
            angles = []
            for text in self._texts:
                orig_rot = text.get_rotation()
                orig_visible = text.get_visible()
                # Reset rotation and visibility
                # for correct window extent calculation
                text.set_rotation(0)
                text.set_visible(True)
                word_px_width = text.get_window_extent(renderer).width
                # In some Matplotlib versions the window extent of
                # whitespace characters is 'nan'
                # In this case, assign a fixed width
                if np.isnan(word_px_width):
                    word_px_width = 5.0
                word_angle \
                    = 2*np.pi * word_px_width / circle_px_circumference
                angles.append(word_angle)
                # Restore
                text.set_rotation(orig_rot)
                text.set_visible(orig_visible)
            return angles
        

        def draw(self, renderer, *args, **kwargs):
            angles = self.get_word_angles(renderer)
            total_angle = np.sum(angles)

            rad_angle = 360 - np.rad2deg(self._angle)
            # Avoid to draw the text upside down, when drawn on the
            # bottom half of the map
            if rad_angle > 90 and rad_angle < 270:
                turn_around = True
            else:
                turn_around = False
            
            # Now that the angle for each word is known,
            # the appropriate position and rotation can be set
            if turn_around:
                # curr_angle is the left-aligned position of the
                # upcoming word
                curr_angle = self._angle + total_angle / 2
            else:
                curr_angle = self._angle - total_angle / 2
            for text, angle in zip(self._texts, angles):
                if turn_around:
                    # The text itself is centered
                    # -> The position itself must be corrected with
                    # half of the word angle
                    angle_corrected = curr_angle - angle / 2
                    text_rot = 360 - np.rad2deg(angle_corrected) + 180
                    curr_angle -= angle
                else:
                    angle_corrected = curr_angle + angle / 2
                    text_rot = 360 - np.rad2deg(angle_corrected)
                    curr_angle += angle
                text.set_position((angle_corrected, self._radius))
                text.set_rotation(text_rot)


except ImportError:
    pass




def _loc_to_rad(loc, plasmid_size):
    if loc > plasmid_size:
        raise ValueError(
            f"Location {loc} is larger then the plasmid size of {plasmid_size}"
        )
    # Location starts at 1 -> (loc-1)
    return ((loc-1) / plasmid_size) * 2*np.pi


def _rad_to_loc(rad, plasmid_size):
    # Location starts at 1 -> + 1
    return rad / (2*np.pi) * plasmid_size + 1


def _merge_over_periodic_boundary(feature, plasmid_size):
    if len(feature.locs) == 1:
        # Only one location -> no merge possible
        return feature
    first_loc = None
    last_loc  = None
    # Find total first location of the feature
    for loc in feature.locs:
        if first_loc is None or loc.first < first_loc.first:
            first_loc = loc
    # Find total last location of the feature
    for loc in feature.locs:
        if last_loc is None or loc.last > last_loc.last:
            last_loc = loc
    # If the first and last location meet at the periodic boundary of
    # the plasmid -> merge them
    if first_loc.first == 1 and last_loc.last == plasmid_size \
        and first_loc.strand == last_loc.strand:
            new_locs = set(feature.locs)
            new_locs.remove(first_loc)
            new_locs.remove(last_loc)
            new_locs.add(Location(
                # the fist base is now at negative location
                # by shifting by one plasmid 'period'
                first  = last_loc.first - plasmid_size,
                last   = first_loc.last,
                strand = first_loc.strand,
                defect = first_loc.defect | last_loc.defect
            ))
            return Feature(feature.key, new_locs, feature.qual)
    else:
        return feature


# ' ', '-' and '_' are word delimiters
separators = re.compile(r"\s|_|-")
def _split_into_words(string):
    match_indices = sorted(
        [match.start() for match in separators.finditer(string)]
    )
    current_index = 0
    words = []
    for i in match_indices:
        # Add word up to delimiter
        words.append(string[current_index : i])
        # Add delimiter
        words.append(string[i : i+1])
        current_index = i+1
    # If there is a word after the last delimiter, add it too
    if current_index < len(string):
        words.append(string[current_index:])
    return words


def _default_feature_formatter(f):
    """
    Returns
    -------
    directional : bool
        True, if the direction of the feature should be indicated by
        an arrow.
    face_color: tuple or str, optional
        A matplotlib compatible color for the feature indicator.
    label_color: tuple or str, optional
        A matplotlib compatible color for the feature label.
    label: str or None
        The label to be displayed for this feature.
        None, if no label should be displayed.
    """
    # Source
    if f.key == "source":
        if f.qual.get("organism") is not None:
            label = f"Source: {f.qual.get('organism')}"
        else:
            label = None
        return False, "black", "white", label
    
    # Origin of Replication
    elif f.key == "rep_origin":
        return False, "indigo", "white", \
               f.qual.get("standard_name", "ori")
    
    # Coding sequences
    elif f.key in ["gene", "CDS", "rRNA"]:
        label = f.qual.get("product")
        if label is None:
            label = f.qual.get("gene")
        return True, colors["orange"], "black", label
    
    elif f.key == "regulatory":
        # Promoters
        if f.qual.get("regulatory_class") in [
            "promoter",
            "TATA_box",
            "minus_35_signal",
            "minus_10_signal"
        ]:
            return True, colors["dimgreen"], "black", f.qual.get("note")
        
        # Terminators
        elif f.qual.get("regulatory_class") in "terminator":
            return False, "firebrick", "white", f.qual.get("note")
        
        # RBS
        elif f.qual.get("regulatory_class") == "ribosome_binding_site":
            return False, colors["brightorange"], "white", None
    
    # Primers
    elif f.key == "primer_bind":
        return True, "royalblue", "black", f.qual.get("note")
    
    # Binding proteins
    elif f.key == "protein_bind":
        return False, colors["lightgreen"], "black", f.qual.get("note")
    
    # Misc
    return True, "dimgray", "white", f.qual.get("note")


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Functions for converting a sequence from/to a GenBank file.
"""

__name__ = "biotite.sequence.io.genbank"
__author__ = "Patrick Kunzmann"
__all__ = ["get_raw_sequence", "get_sequence", "get_annotated_sequence",
           "set_sequence", "set_annotated_sequence"]

import re
from ....file import InvalidFileError
from ...seqtypes import ProteinSequence, NucleotideSequence
from ...annotation import AnnotatedSequence
from .file import GenBankFile
from .annotation import get_annotation, set_annotation


_SYMBOLS_PER_CHUNK = 10
_SEQ_CHUNKS_PER_LINE = 6
_SYMBOLS_PER_LINE = _SYMBOLS_PER_CHUNK * _SEQ_CHUNKS_PER_LINE


def get_raw_sequence(gb_file):
    """
    Get the raw sequence string from the *ORIGIN* field
    of a GenBank file.

    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *ORIGIN* field from.
    
    Returns
    -------
    seq_str: str
        The unaltered sequence as string.
        Sequence positions and whitespace characters are removed.
    """
    fields = gb_file.get_fields("ORIGIN")
    if len(fields) == 0:
        raise InvalidFileError("File has no 'ORIGIN' field")
    if len(fields) > 1:
        raise InvalidFileError("File has multiple 'ORIGIN' fields")
    lines, _ = fields[0]
    return _field_to_seq_string(lines)


def get_sequence(gb_file, format="gb"):
    """
    Get the sequence from the *ORIGIN* field of a GenBank file.

    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the *ORIGIN* field from.
    format : {'gb', 'gp'}
        Indicates whether the file is a GenBank or a GenPept file.
        Depending on this parameter a :class:`NucleotideSequence` or a
        :class:`ProteinSequence` is returned.
    
    Returns
    -------
    sequence : NucleotideSequence or ProteinSequence
        The reference sequence in the file.
    """
    return _convert_seq_str(get_raw_sequence(gb_file), format)


def get_annotated_sequence(gb_file, format="gb", include_only=None):
    """
    Get an annotated sequence by combining the *ANNOTATION* and
    *ORIGIN* fields of a GenBank file.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to read the fields from.
    include_only : iterable object of str, optional
        List of names of feature keys, which should included
        in the annotation. By default all features are included.
    
    Returns
    ----------
    annot_seq : AnnotatedSequence
        The annotated sequence.
    """
    fields = gb_file.get_fields("ORIGIN")
    if len(fields) == 0:
        raise InvalidFileError("File has no 'ORIGIN' field")
    if len(fields) > 1:
        raise InvalidFileError("File has multiple 'ORIGIN' fields")
    lines, _ = fields[0]
    sequence = _convert_seq_str(_field_to_seq_string(lines), format)
    seq_start = _get_seq_start(lines)
    annotation = get_annotation(gb_file, include_only)
    return AnnotatedSequence(annotation, sequence, sequence_start=seq_start)


def _field_to_seq_string(origin_content):
    seq_str = "".join(origin_content)
    # Remove numbers and emtpy spaces
    regex = re.compile("[0-9]| ")
    seq_str = regex.sub("", seq_str)
    return seq_str


def _convert_seq_str(seq_str, format):
    if len(seq_str) == 0:
        raise InvalidFileError("The file's 'ORIGIN' field is empty")
    if format == "gb":
        return NucleotideSequence(seq_str.replace("U","T").replace("X","N"))
    elif format == "gp":
        if "U" in seq_str:
            warnings.warn(
                "ProteinSequence objects do not support selenocysteine (U), "
                "occurrences were substituted by cysteine (C)"
            )
            seq_str = seq_str.replace("U", "C")
        return ProteinSequence(seq_str)
    else:
        raise ValueError(f"Unknown format '{format}'")
    

def _get_seq_start(origin_content):
    # Start of sequence is the sequence position indicator
    # at the beginning of the first line
    return int(origin_content[0].split()[0])




def set_sequence(gb_file, sequence, sequence_start=1):
    """
    Set the *ORIGIN* field of a GenBank file with a sequence.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to be edited.
    sequence : str or NucleotideSequence or ProteinSequence
        The sequence that is put into the GenBank file.
    sequence_start : int, optional
        The number of the first base of the sequence.
    """
    lines = []
    seq_str = str(sequence).lower()
    line = "{:>9d}".format(sequence_start)
    for i in range(0, len(sequence), _SYMBOLS_PER_CHUNK):
        # New line after 5 sequence chunks
        if i != 0 and i % _SYMBOLS_PER_LINE == 0:
            lines.append(line)
            line = "{:>9d}".format(sequence_start + i)
        line += " " + str(seq_str[i : i + _SYMBOLS_PER_CHUNK])
    # Append last line
    lines.append(line)
    gb_file.set_field("ORIGIN", lines)


def set_annotated_sequence(gb_file, annot_sequence):
    """
    Set the *FEATURES* and *ORIGIN* fields of a GenBank file with the
    annotation and sequence of an annotated sequence.
    
    Parameters
    ----------
    gb_file : GenBankFile
        The GenBank file to be edited.
    annot_sequence : AnnotatedSequence
        The annotated sequence that is put into the GenBank file.
    """
    set_annotation(gb_file, annot_sequence.annotation)
    set_sequence(
        gb_file, annot_sequence.sequence, annot_sequence.sequence_start
    )


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for visualizing structure related objects.
"""

__name__ = "biotite.structure.graphics"
__author__ = "Patrick Kunzmann, Tom David Müller"

from .atoms import *
from .rna import *



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.structure as struc
import biotite.structure.io as strucio
import numpy as np
from os.path import join
from ..util import data_dir
import pytest


@pytest.fixture
def array():
    return strucio.load_structure(join(data_dir("structure"), "1igy.mmtf"))

def test_get_chain_starts(array):
    _, ref_starts = np.unique(array.chain_id, return_index=True)
    test_starts = struc.get_chain_starts(array)
    # All first occurences of a chain id are automatically chain starts
    assert set(ref_starts).issubset(set(test_starts))

def test_get_chains(array):
    assert struc.get_chains(array).tolist() == ["A", "B", "C", "D", "E", "F"]

def test_get_chain_count(array):
    assert struc.get_chain_count(array) == 6

def test_chain_iter(array):
    n = 0
    for chain in struc.get_chains(array):
        n += 1
        assert isinstance(array, struc.AtomArray)
    assert n == 6


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from biotite.sequence import NucleotideSequence
from biotite.sequence import ProteinSequence
from biotite.sequence import Alphabet
from biotite.sequence import GeneralSequence
from biotite.sequence import LetterAlphabet
from biotite.sequence import Location
from biotite.sequence import Feature
from biotite.sequence import Annotation
from biotite.sequence import AnnotatedSequence
from biotite.sequence.align import Alignment
from biotite.structure import Atom
import numpy as np
from numpy import float32, int32
from biotite.sequence import CodonTable
from biotite.sequence.align import SubstitutionMatrix
from biotite.sequence import SequenceProfile
import pytest

__author__ = "Maximilian Greil"


@pytest.mark.parametrize("repr_object",
                         [NucleotideSequence("AACTGCTA"),
                          NucleotideSequence("AACTGCTA", ambiguous=True),
                          ProteinSequence("BIQTITE"),
                          Alphabet(["X", "Y", "Z"]),
                          GeneralSequence(Alphabet(["X", 42, False]), ["X", 42, "X"]),
                          LetterAlphabet(["X", "Y", "Z"]),
                          Location(98, 178),
                          Feature("CDS", [Location(98, 178)], qual={"gene": "test1"}),
                          Annotation([Feature("CDS", [Location(98, 178)], qual={"gene": "test1"})]),
                          AnnotatedSequence(Annotation([Feature("CDS", [Location(98, 178)], qual={"gene": "test1"})]),
                                            NucleotideSequence("AACTGCTA")),
                          Alignment([NucleotideSequence("CGTCAT", ambiguous=False),
                                     NucleotideSequence("TCATGC", ambiguous=False)],
                                    np.array([[0, -1], [1, -1], [2, 0], [3, 1], [4, 2], [5, 3], [-1, 4], [-1, 5]]),
                                    score=-20),
                          Atom([1, 2, 3], chain_id="A"),
                          CodonTable.default_table(),
                          SubstitutionMatrix(Alphabet(["foo", "bar"]), Alphabet([1, 2, 3]),
                                             {("foo", 1): 5, ("foo", 2): 10, ("foo", 3): 15, ("bar", 1): 42,
                                              ("bar", 2): 42, ("bar", 3): 42}),
                          SequenceProfile(np.array([[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 2], [0, 2, 0, 0],
                                                    [2, 0, 0, 0], [0, 0, 0, 2], [0, 0, 1, 0], [0, 1, 0, 0]]),
                                          np.array([1, 1, 0, 0, 0, 0, 1, 1]),
                                          Alphabet(["A", "C", "G", "T"]))])
def test_repr(repr_object):
    assert eval(repr(repr_object)) == repr_object



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.mafft"
__author__ = "Patrick Kunzmann"
__all__ = ["MafftApp"]

import re
import os
from ..msaapp import MSAApp
from ..application import AppState, requires_state
from ...sequence.sequence import Sequence
from ...sequence.seqtypes import NucleotideSequence, ProteinSequence
from ...sequence.io.fasta.file import FastaFile
from ...sequence.align.alignment import Alignment
from ...sequence.phylo.tree import Tree


_prefix_pattern = re.compile("._")



class MafftApp(MSAApp):
    """
    Perform a multiple sequence alignment using MAFFT.
    
    Parameters
    ----------
    sequences : list of Sequence
        The sequences to be aligned.
    bin_path : str, optional
        Path of the MUSCLE binary.
    matrix : SubstitutionMatrix, optional
        A custom substitution matrix.
    
    Examples
    --------

    >>> seq1 = ProteinSequence("BIQTITE")
    >>> seq2 = ProteinSequence("TITANITE")
    >>> seq3 = ProteinSequence("BISMITE")
    >>> seq4 = ProteinSequence("IQLITE")
    >>> app = MafftApp([seq1, seq2, seq3, seq4])
    >>> app.start()
    >>> app.join()
    >>> alignment = app.get_alignment()
    >>> print(alignment)
    -BIQTITE
    TITANITE
    -BISMITE
    --IQLITE
    """
    
    def __init__(self, sequences, bin_path="mafft", matrix=None):
        super().__init__(sequences, bin_path, matrix)
        self._tree = None
        self._out_tree_file_name = self.get_input_file_path() + ".tree"
    
    def run(self):
        args = [
            "--quiet",
            "--auto",
            "--treeout",
            # Get the reordered alignment in order for
            # get_alignment_order() to work properly 
            "--reorder",
        ]
        if self.get_seqtype() == "protein":
            args += ["--amino"]
        else:
            args += ["--nuc"]
        if self.get_matrix_file_path() is not None:
            args += ["--aamatrix", self.get_matrix_file_path()]
        args += [self.get_input_file_path()]
        self.set_arguments(args)
        super().run()
    
    def evaluate(self):
        with open(self.get_output_file_path(), "w") as f:
            # MAFFT outputs alignment to stdout
            # -> write stdout to output file name
            f.write(self.get_stdout())
        super().evaluate()
        with open(self._out_tree_file_name, "r") as file:
            raw_newick = file.read().replace("\n", "")
            # Mafft uses sequences label in the form '<n>_<seqname>'
            # Only the <seqname> is required
            # -> remove the '<n>_' prefix
            newick = re.sub(_prefix_pattern, "", raw_newick)
            self._tree = Tree.from_newick(newick)
    
    def clean_up(self):
        os.remove(self._out_tree_file_name)

    @requires_state(AppState.JOINED)
    def get_guide_tree(self):
        """
        Get the guide tree created for the progressive alignment.
        
        Returns
        -------
        tree : Tree
            The guide tree.
        """
        return self._tree
    
    @staticmethod
    def supports_nucleotide():
        return True
    
    @staticmethod
    def supports_protein():
        return True
    
    @staticmethod
    def supports_custom_nucleotide_matrix():
        return True
    
    @staticmethod
    def supports_custom_protein_matrix():
        return True



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import biotite.sequence as seq
import numpy as np
import pytest


def test_find_subsequence():
    string = "ATACGCTTGCT"
    substring = "GCT"
    main_seq = seq.NucleotideSequence(string)
    sub_seq = seq.NucleotideSequence(substring)
    matches = seq.find_subsequence(main_seq, sub_seq)
    assert list(matches) == [4,8]
    
def test_find_symbol():
    string = "ATACGCTTGCT"
    symbol = "T"
    dna = seq.NucleotideSequence(string)
    assert list(seq.find_symbol(dna, symbol)) == [1,6,7,10]
    assert seq.find_symbol_first(dna, symbol) == 1
    assert seq.find_symbol_last(dna, symbol) == 10


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.muscle"
__author__ = "Patrick Kunzmann"
__all__ = ["MuscleApp"]

import re
import numbers
import warnings
import subprocess
from tempfile import NamedTemporaryFile
from ..localapp import cleanup_tempfile
from ..msaapp import MSAApp
from ..application import AppState, VersionError, requires_state
from ...sequence.sequence import Sequence
from ...sequence.seqtypes import NucleotideSequence, ProteinSequence
from ...sequence.align.matrix import SubstitutionMatrix
from ...sequence.align.alignment import Alignment
from ...sequence.phylo.tree import Tree


class MuscleApp(MSAApp):
    """
    Perform a multiple sequence alignment using MUSCLE version 3.
    
    Parameters
    ----------
    sequences : list of Sequence
        The sequences to be aligned.
    bin_path : str, optional
        Path of the MUSCLE binary.
    matrix : SubstitutionMatrix, optional
        A custom substitution matrix.
    
    See also
    ---------
    Muscle5App
    
    Examples
    --------

    >>> seq1 = ProteinSequence("BIQTITE")
    >>> seq2 = ProteinSequence("TITANITE")
    >>> seq3 = ProteinSequence("BISMITE")
    >>> seq4 = ProteinSequence("IQLITE")
    >>> app = MuscleApp([seq1, seq2, seq3, seq4])
    >>> app.start()
    >>> app.join()
    >>> alignment = app.get_alignment()
    >>> print(alignment)
    BIQT-ITE
    TITANITE
    BISM-ITE
    -IQL-ITE
    """
    
    def __init__(self, sequences, bin_path="muscle", matrix=None):
        major_version = get_version(bin_path)[0]
        if major_version != 3:
            raise VersionError(
                f"Muscle 3 is required, got version {major_version}"
            )
        
        super().__init__(sequences, bin_path, matrix)
        self._gap_open = None
        self._gap_ext = None
        self._terminal_penalty = None
        self._tree1 = None
        self._tree2 = None
        self._out_tree1_file = NamedTemporaryFile(
            "r", suffix=".tree", delete=False
        )
        self._out_tree2_file = NamedTemporaryFile(
            "r", suffix=".tree", delete=False
        )
    
    def run(self):
        args = [
            "-quiet",
            "-in", self.get_input_file_path(),
            "-out", self.get_output_file_path(),
            "-tree1", self._out_tree1_file.name,
            "-tree2", self._out_tree2_file.name,
        ]
        if self.get_seqtype() == "protein":
            args += ["-seqtype", "protein"]
        else:
            args += ["-seqtype", "dna"]
        if self.get_matrix_file_path() is not None:
            args += ["-matrix", self.get_matrix_file_path()]
        if self._gap_open is not None and self._gap_ext is not None:
            args += ["-gapopen",   f"{self._gap_open:.1f}"]
            args += ["-gapextend", f"{self._gap_ext:.1f}"]
            # When the gap penalty is set,
            # use the penalty also for hydrophobic regions
            args += ["-hydrofactor", "1.0"]
            # Use the recommendation of the documentation
            args += ["-center", "0.0"]
        self.set_arguments(args)
        super().run()
    
    def evaluate(self):
        super().evaluate()

        newick = self._out_tree1_file.read().replace("\n", "")
        if len(newick) > 0:
            self._tree1 = Tree.from_newick(newick)
        else:
            warnings.warn(
                "MUSCLE did not write a tree file from the first iteration"
            )
        
        newick = self._out_tree2_file.read().replace("\n", "")
        if len(newick) > 0:
            self._tree2 = Tree.from_newick(newick)
        else:
            warnings.warn(
                "MUSCLE did not write a tree file from the second iteration"
            )
    
    def clean_up(self):
        super().clean_up()
        cleanup_tempfile(self._out_tree1_file)
        cleanup_tempfile(self._out_tree2_file)
    
    @requires_state(AppState.CREATED)
    def set_gap_penalty(self, gap_penalty):
        """
        Set the gap penalty for the alignment.

        Parameters
        ----------
        gap_penalty : float or (tuple, dtype=int)
            If a float is provided, the value will be interpreted as
            general gap penalty.
            If a tuple is provided, an affine gap penalty is used.
            The first value in the tuple is the gap opening penalty,
            the second value is the gap extension penalty.
            The values need to be negative.
        """
        # Check if gap penalty is general or affine
        if isinstance(gap_penalty, numbers.Real):
            if gap_penalty > 0:
                raise ValueError("Gap penalty must be negative")
            self._gap_open = gap_penalty
            self._gap_ext= gap_penalty
        elif type(gap_penalty) == tuple:
            if gap_penalty[0] > 0 or gap_penalty[1] > 0:
                    raise ValueError("Gap penalty must be negative")
            self._gap_open = gap_penalty[0]
            self._gap_ext = gap_penalty[1]
        else:
            raise TypeError("Gap penalty must be either float or tuple")
    
    @requires_state(AppState.JOINED)
    def get_guide_tree(self, iteration="identity"):
        """
        Get the guide tree created for the progressive alignment.
        
        Parameters
        ----------
        iteration : {'kmer', 'identity'}
            If 'kmer', the first iteration tree is returned.
            This tree uses the sequences common *k*-mers as distance
            measure.
            If 'identity' the second iteration tree is returned.
            This tree uses distances based on the pairwise sequence
            identity after the first progressive alignment iteration.
        
        Returns
        -------
        tree : Tree
            The guide tree.
        """
        if iteration == "kmer":
            return self._tree1
        elif iteration == "identity":
            return self._tree2
        else:
            raise ValueError("Iteration must be 'kmer' or 'identity'")
    
    @staticmethod
    def supports_nucleotide():
        return True
    
    @staticmethod
    def supports_protein():
        return True
    
    @staticmethod
    def supports_custom_nucleotide_matrix():
        return False
    
    @staticmethod
    def supports_custom_protein_matrix():
        return True
    
    @classmethod
    def align(cls, sequences, bin_path=None, matrix=None,
              gap_penalty=None):
        """
        Perform a multiple sequence alignment.
        
        This is a convenience function, that wraps the :class:`MuscleApp`
        execution.
        
        Parameters
        ----------
        sequences : iterable object of Sequence
            The sequences to be aligned
        bin_path : str, optional
            Path of the MSA software binary. By default, the default path
            will be used.
        matrix : SubstitutionMatrix, optional
            A custom substitution matrix.
        gap_penalty : float or (tuple, dtype=int), optional
            If a float is provided, the value will be interpreted as
            general gap penalty.
            If a tuple is provided, an affine gap penalty is used.
            The first value in the tuple is the gap opening penalty,
            the second value is the gap extension penalty.
            The values need to be negative.
        
        Returns
        -------
        alignment : Alignment
            The global multiple sequence alignment.
        """
        if bin_path is None:
            app = cls(sequences, matrix=matrix)
        else:
            app = cls(sequences, bin_path, matrix=matrix)
        if gap_penalty is not None:
            app.set_gap_penalty(gap_penalty)
        app.start()
        app.join()
        return app.get_alignment()


def get_version(bin_path="muscle"):
    output = subprocess.run(
        [bin_path, "-version"], capture_output=True, text=True
    )
    # Find matches for version string containing major and minor version
    match = re.search("\d+\.\d+", output.stdout)
    if match is None:
        raise subprocess.SubprocessError("Application did not print version")
    version_string = match.group(0)
    splitted = version_string.split(".")
    return int(splitted[0]), int(splitted[1])


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.info as info


@pytest.fixture
def array():
    """
    Create an :class:`AtomArray` containing a lot of different
    molecules.
    The atoms that belong to a single molecule are not adjacent in the
    :class:`AtomArray`, but a are shuffled in random positions of the
    :class:`AtomArray`.
    """
    MOL_NAMES = [
        "ARG", # Molecule with multiple branches
        "TRP", # Molecule with a cycle
        "GLC", # Molecule with a cycle
        "NA",  # A single atom
        "ATP"  # Larger molecule
    ]
    N_MOLECULES = 20

    np.random.seed(0)
    
    atom_array = struc.AtomArray(0)
    for i, mol_name in enumerate(np.random.choice(MOL_NAMES, N_MOLECULES)):
        molecule = info.residue(mol_name)
        molecule.res_id[:] = i+1
        atom_array += molecule
    
    reordered_indices = np.random.choice(
        np.arange(atom_array.array_length()),
        atom_array.array_length(),
        replace=False
    )
    atom_array = atom_array[reordered_indices]

    return atom_array


@pytest.mark.parametrize(
    "as_stack, as_bonds",
    [
        (False, False),
        (True,  False),
        (False, True )
    ]
)
def test_get_molecule_indices(array, as_stack, as_bonds):
    """
    Multiple tests to :func:`get_molecule_indices()` on a
    :class:`AtomArray` of random molecules.
    """
    if as_stack:
        array = struc.stack([array])
    
    if as_bonds:
        test_indices = struc.get_molecule_indices(array.bonds)
    else:
        test_indices = struc.get_molecule_indices(array)
    
    seen_atoms = 0
    for indices in test_indices:
        molecule = array[..., indices]
        # Assert that all residue IDs in the molecule are equal
        # -> all atoms from the same molecule
        assert (molecule.res_id == molecule.res_id[0]).all()
        # Assert that no atom is missing from the molecule
        assert molecule.array_length() \
            == info.residue(molecule.res_name[0]).array_length()
        seen_atoms += molecule.array_length()
    # Assert that all molecules are fond
    assert seen_atoms == array.array_length()


@pytest.mark.parametrize(
    "as_stack, as_bonds",
    [
        (False, False),
        (True,  False),
        (False, True )
    ]
)
def test_get_molecule_masks(array, as_stack, as_bonds):
    """
    Test whether the masks returned by :func:`get_molecule_masks()`
    point to the same atoms as the indices returned by
    :func:`get_molecule_indices()`.
    """
    if as_stack:
        array = struc.stack([array])
    
    if as_bonds:
        ref_indices = struc.get_molecule_indices(array.bonds)
        test_masks = struc.get_molecule_masks(array.bonds)
    else:
        ref_indices = struc.get_molecule_indices(array)
        test_masks = struc.get_molecule_masks(array)
    
    for i in range(len(test_masks)):
        # Assert that the mask is 'True' for all indices
        # and that these 'True' values are the only ones in the mask
        assert (test_masks[i, ref_indices[i]] == True).all()
        assert np.count_nonzero(test_masks[i]) == len(ref_indices[i])


@pytest.mark.parametrize("as_stack", (False, True))
def test_molecule_iter(array, as_stack):
    """
    Test whether :func:`molecule_iter()` gives the same molecules as
    pointed by :func:`get_molecule_indices()`.
    """
    if as_stack:
        array = struc.stack([array])

    ref_indices = struc.get_molecule_indices(array)
    test_iterator = struc.molecule_iter(array)

    for i, molecule in enumerate(test_iterator):
        assert molecule == array[..., ref_indices[i]]


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.mol"
__author__ = "Patrick Kunzmann"
__all__ = ["get_structure", "set_structure"]



def get_structure(mol_file):
    """
    Get an :class:`AtomArray` from the MOL file.

    Ths function is a thin wrapper around
    :meth:`MOLFile.get_structure()`.

    Parameters
    ----------
    mol_file : MOLFile
        The MOL file.
    
    Returns
    -------
    array : AtomArray
        This :class:`AtomArray` contains the optional ``charge``
        annotation and has an associated :class:`BondList`.
        All other annotation categories, except ``element`` are
        empty.
    """
    return mol_file.get_structure()
    

def set_structure(mol_file, atoms):
    """
    Set the :class:`AtomArray` for the MOL file.

    Ths function is a thin wrapper around
    :meth:`MOLFile.set_structure()`.
    
    Parameters
    ----------
    mol_file : MOLFile
        The MOL file.
    array : AtomArray
        The array to be saved into this file.
        Must have an associated :class:`BondList`.
    """
    mol_file.set_structure(atoms)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing an :class:`AtomArray` or
:class:`AtomArrayStack` using the internal NPZ file format. This binary
format is used to store `NumPy` arrays. Since atom arrays and stacks are
completely built on `NumPy` arrays, this format is preferable for
Biotite internal usage due to fast I/O operations and preservation
of all atom annotation arrays.
"""

__name__ = "biotite.structure.io.npz"
__author__ = "Patrick Kunzmann"

from .file import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import TemporaryFile
import glob
from os.path import join
import biotite.sequence as seq
import biotite.sequence.io.genbank as gb
import numpy as np
import pytest
from ..util import data_dir


@pytest.mark.parametrize(
    "path",
    glob.glob(join(data_dir("sequence"), "*.gb")) + \
    glob.glob(join(data_dir("sequence"), "[!multifile]*.gp"))
)
def test_contiguous_field_pos(path):
    """
    Check whether the internal index of a GenBankFile is contiguous
    """
    gb_file = gb.GenBankFile.read(path)
    assert gb_file._field_pos[0][0] == 0
    for i in range(1, len(gb_file._field_pos)):
        start, _, _ = gb_file._field_pos[i]
        _, stop, _  = gb_file._field_pos[i-1]
        assert start == stop


def test_file_access():
    """
    Test getting, setting, deleting and inserting fields in a GenBank
    file.
    """
    gb_file = gb.GenBankFile()
    gb_file.append("SOMEFIELD", ["Some content", "some other content"])
    gb_file.insert(0, "OTHERFIELD", ["Additional content"])
    assert gb_file[1] \
        == ("SOMEFIELD", ["Some content", "some other content"], {})
    gb_file[1] \
        = "NEWFIELD", ["Extra content"], {"SUBFIELD" : ["L 1", "L 2"]}
    gb_file.append("THIRDFIELD", ["Supplementary content"])
    assert len(gb_file) == 3
    assert gb_file[0] == ("OTHERFIELD", ["Additional content"], {})
    del gb_file[0]
    assert gb_file[0] \
        == ("NEWFIELD", ["Extra content"], {"SUBFIELD" : ["L 1", "L 2"]})
    del gb_file[0]
    assert gb_file[0] == ("THIRDFIELD", ["Supplementary content"], {})
    del gb_file[0]
    assert len(gb_file) == 0



@pytest.mark.parametrize(
    "path",
    glob.glob(join(data_dir("sequence"), "*.gb")) + \
    glob.glob(join(data_dir("sequence"), "[!multifile]*.gp"))
)
def test_conversion_lowlevel(path):
    """
    Test whether the low-level GenBank interface can properly read
    a GenBank file and write a file, without data changing.
    """
    gb_file = gb.GenBankFile.read(path)
    ref_parsed_fields = [field for field in gb_file]

    gb_file = gb.GenBankFile()
    for name, content, subfields in ref_parsed_fields:
        gb_file.append(name, content, subfields)
    temp = TemporaryFile("w+")
    gb_file.write(temp)
    
    temp.seek(0)
    gb_file = gb.GenBankFile.read(temp)
    temp.close()
    test_parsed_fields = [field for field in gb_file]
    assert test_parsed_fields == ref_parsed_fields


@pytest.mark.parametrize(
    "path",
    glob.glob(join(data_dir("sequence"), "*.gb")) + \
    glob.glob(join(data_dir("sequence"), "[!multifile]*.gp"))
)
def test_conversion_highlevel(path):
    """
    Test whether the high-level GenBank interface can properly read
    the locus, annotation and sequence from GenBank file and write
    these properties to a file, without data changing.
    """
    suffix = path[-2:]
    gb_file = gb.GenBankFile.read(path)
    ref_locus = gb.get_locus(gb_file)
    ref_annot_seq = gb.get_annotated_sequence(gb_file, format=suffix)

    gb_file = gb.GenBankFile()
    gb.set_locus(gb_file, *ref_locus)
    gb.set_annotated_sequence(gb_file, ref_annot_seq)
    temp = TemporaryFile("w+")
    gb_file.write(temp)
    
    temp.seek(0)
    gb_file = gb.GenBankFile.read(temp)
    temp.close()
    test_locus = gb.get_locus(gb_file)
    test_annot_seq = gb.get_annotated_sequence(gb_file, format=suffix)
    assert test_locus == ref_locus
    assert test_annot_seq.sequence       == ref_annot_seq.sequence
    assert test_annot_seq.annotation     == ref_annot_seq.annotation
    assert test_annot_seq.sequence_start == ref_annot_seq.sequence_start


def test_genbank_utility_gb():
    """
    Check whether the high-level utility functions return the expected
    content of a known GenBank file. 
    """
    gb_file = gb.GenBankFile.read(join(data_dir("sequence"), "ec_bl21.gb"))
    assert gb.get_locus(gb_file) \
        == ("CP001509", 4558953, "DNA", True, "BCT", "16-FEB-2017")
    assert gb.get_definition(gb_file) \
        == ("Escherichia coli BL21(DE3), complete genome.")
    assert gb.get_version(gb_file) == "CP001509.3"
    assert gb.get_gi(gb_file) == 296142109
    assert gb.get_db_link(gb_file) \
        == {"BioProject" : "PRJNA20713", "BioSample" : "SAMN02603478"}
    annotation = gb.get_annotation(gb_file, include_only=["CDS"])
    feature = seq.Feature(
        "CDS",
        [seq.Location(5681, 6457, seq.Location.Strand.REVERSE)],
        {"gene": "yaaA", "transl_table": "11"}
    )
    in_annotation = False
    for f in annotation:
        if f.key == feature.key and f.locs == feature.locs and \
           all([(key, val in f.qual.items())
                for key, val in feature.qual.items()]):
                    in_annotation = True
    assert in_annotation
    assert len(gb.get_sequence(gb_file, format="gb")) == 4558953


def test_genbank_utility_gp():
    """
    Check whether the high-level utility functions return the expected
    content of a known GenPept file. 
    """
    gp_file = gb.GenBankFile.read(join(data_dir("sequence"), "bt_lysozyme.gp"))
    #[print(e) for e in gp_file._field_pos]
    assert gb.get_locus(gp_file) \
        == ("AAC37312", 147, "", False, "MAM", "27-APR-1993")
    assert gb.get_definition(gp_file) == "lysozyme [Bos taurus]."
    assert gb.get_version(gp_file) == "AAC37312.1"
    assert gb.get_gi(gp_file) == 163334
    annotation = gb.get_annotation(gp_file)
    feature = seq.Feature(
        "Site",
        [seq.Location(start, stop) for start, stop in zip(
            [52,55,62,76,78,81,117,120,125],
            [53,55,62,76,78,81,117,120,126]
        )],
        {"note": "lysozyme catalytic cleft [active]", "site_type": "active"}
    )
    in_annotation = False
    for f in annotation:
        if f.key == feature.key and f.locs == feature.locs and \
           all([(key, val in f.qual.items())
                for key, val in feature.qual.items()]):
                    in_annotation = True
    assert in_annotation
    assert len(gb.get_sequence(gp_file, format="gp")) == 147


def test_multi_file():
    multi_file = gb.MultiFile.read(join(data_dir("sequence"), "multifile.gp"))
    accessions = [gb.get_accession(f) for f in multi_file]
    assert accessions == ["1L2Y_A", "3O5R_A", "5UGO_A"]


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.tantan"
__author__ = "Patrick Kunzmann"
__all__ = ["TantanApp"]

import io
from tempfile import NamedTemporaryFile
import numpy as np
from ..localapp import LocalApp, cleanup_tempfile
from ..application import AppState, requires_state
from ...sequence.sequence import Sequence
from ...sequence.seqtypes import NucleotideSequence, ProteinSequence
from ...sequence.io.fasta.file import FastaFile
from ...sequence.io.fasta.convert import set_sequence
from ..util import map_sequence, map_matrix


MASKING_LETTER = "!"


class TantanApp(LocalApp):
    r"""
    Mask sequence reapeat regions using *tantan*. :footcite:`Frith2011`

    Parameters
    ----------
    sequence : NucleotideSequence or ProteinSequence
        The sequence to be masked.
    matrix : SubstitutionMatrix, optional
        The substitution matrix to use for repeat identification.
        A sequence segment is considered to be a repeat of another
        segment, if the substitution score between these segments is
        greater than a threshold value.
    bin_path : str, optional
        Path of the *tantan* binary.

    References
    ----------
    
    .. footbibliography::

    Examples
    --------

    >>> sequence = NucleotideSequence("GGCATCGATATATATATATAGTCAA")
    >>> app = TantanApp(sequence)
    >>> app.start()
    >>> app.join()
    >>> repeat_mask = app.get_mask()
    >>> print(repeat_mask)
    [False False False False False False False False False  True  True  True
      True  True  True  True  True  True  True  True False False False False
     False]
    >>> print(sequence, "\n" + "".join(["^" if e else " " for e in repeat_mask]))
    GGCATCGATATATATATATAGTCAA 
             ^^^^^^^^^^^         
    """
    
    def __init__(self, sequence, matrix=None, bin_path="tantan"):
        super().__init__(bin_path)

        if not isinstance(sequence, Sequence):
            raise TypeError("A 'Sequence' object is expected")
        
        if matrix is None:
            self._matrix_file = None
        else:
            if not matrix.get_alphabet1().extends(sequence.alphabet):
                raise ValueError(
                    "The sequence's alphabet does not fit the matrix"
                )
            if not matrix.is_symmetric():
                raise ValueError("A symmetric matrix is required")
            self._matrix_file = NamedTemporaryFile(
                "w", suffix=".mat", delete=False
            )
        
        if isinstance(sequence, NucleotideSequence):
            self._is_protein=False
            self._sequence = sequence
            self._matrix = matrix
        elif isinstance(sequence, ProteinSequence):
            self._is_protein=True
        else:
            raise TypeError(
                "A NucleotideSequence or ProteinSequence is required"
            )
        self._sequence = sequence
        self._matrix = matrix
        
        self._in_file = NamedTemporaryFile("w", suffix=".fa", delete=False)


    def run(self):
        sequence_file = FastaFile()
        set_sequence(sequence_file, self._sequence)
        sequence_file.write(self._in_file)
        self._in_file.flush()
        if self._matrix is not None:
            self._matrix_file.write(str(self._matrix))
            self._matrix_file.flush()
        
        args = []
        if self._matrix is not None:
            args += ["-m", self._matrix_file.name]
        if self._is_protein:
             args += ["-p"]
        args += [
            "-x", MASKING_LETTER,
            self._in_file.name
        ]
        self.set_arguments(args)
        super().run()
    

    def evaluate(self):
        super().evaluate()
        masked_file = FastaFile.read(io.StringIO(self.get_stdout()))
        masked_sequence = masked_file["sequence"]
        array = np.frombuffer(masked_sequence.encode("ASCII"), dtype=np.ubyte)
        self._mask = (array == MASKING_LETTER.encode("ASCII")[0])
    

    def clean_up(self):
        super().clean_up()
        cleanup_tempfile(self._in_file)
        if self._matrix_file is not None:
            cleanup_tempfile(self._matrix_file)
    

    @requires_state(AppState.JOINED)
    def get_mask(self):
        """
        Get a boolean mask covering identified repeat regions of the
        input sequence.

        Returns
        -------
        repeat_mask : ndarray, shape=(n,), dtype=bool
            A boolean mask that is true for each sequence position that
            is identified as repeat.
        """
        return self._mask


    @staticmethod
    def mask_repeats(sequence, matrix=None, bin_path="tantan"):
        """
        Mask repeat regions of the given input sequence.

        Parameters
        ----------
        sequence : NucleotideSequence or ProteinSequence
            The sequence to be masked.
        matrix : SubstitutionMatrix, optional
            The substitution matrix to use for repeat identification.
            A sequence segment is considered to be a repeat of another
            segment, if the substitution score between these segments is
            greater than a threshold value.
        bin_path : str, optional
            Path of the *tantan* binary.

        Returns
        -------
        repeat_mask : ndarray, shape=(n,), dtype=bool
            A boolean mask that is true for each sequence position that
            is identified as repeat.
        """
        app = TantanApp(sequence, matrix, bin_path)
        app.start()
        app.join()
        return app.get_mask()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database.rcsb"
__author__ = "Patrick Kunzmann, Maximilian Dombrowsky"
__all__ = ["Query", "SingleQuery", "CompositeQuery",
           "BasicQuery", "FieldQuery",
           "SequenceQuery", "StructureQuery", "MotifQuery",
           "search", "count"]

import abc
import json
import copy
from datetime import datetime
import numpy as np
import requests
from ...sequence.seqtypes import ProteinSequence, NucleotideSequence
from ..error import RequestError


_search_url = "http://search.rcsb.org/rcsbsearch/v1/query"
_scope_to_target = {
    "protein": "pdb_protein_sequence",
    "rna":     "pdb_rna_sequence",
    "dna":     "pdb_dna_sequence"
}


class Query(metaclass=abc.ABCMeta):
    """
    A representation of a JSON query for the RCSB search API.
    
    This is the abstract base class for all queries.
    """
    @abc.abstractmethod
    def get_content(self):
        """
        Get the query content, i.e. the data belonging to the
        ``'query'`` attribute in the RCSB search API.

        This content is converted into JSON by the :func:`search`
        and :func:`count` methods.
        """
        pass

    def __and__(self, query):
        return CompositeQuery([self, query], "and")

    def __or__(self, query):
        return CompositeQuery([self, query], "or")



class SingleQuery(Query, metaclass=abc.ABCMeta):
    """
    A terminal query node for the RCSB search API.
    
    Multiple :class:`SingleQuery` objects can be combined to
    :class:`CompositeQuery`objects using the ``|`` and ``&`` operators.

    This is the abstract base class for all queries that are
    terminal nodes.
    """
    # The node ID is incremented for every created 'SingleQuery' object
    _node_id = 0

    def __init__(self):
        self._node_id = SingleQuery._node_id
        SingleQuery._node_id += 1

    @abc.abstractmethod
    def get_content(self):
        return {
            "node_id": self._node_id,
            "parameters": {}
        }


class CompositeQuery(Query):
    """
    A group query node for the RCSB search API.
    
    A composite query is an combination of other queries, combined
    either with the `'and'` or `'or'` operator.
    Usually, a :class:`CompositeQuery` will not be created by calling
    its constructor, but by combining queries using the ``|`` or ``&``
    operator.

    Parameters
    ----------
    queries : iterable object of Query
        The queries to be combined.
    operator : {'or', 'and'}
        The type of combination.
    """
    def __init__(self, queries, operator):
        self._queries = queries
        if operator not in ("or", "and"):
            raise ValueError(
                f"Operator must be 'or' or 'and', not '{operator}'"
            )
        self._operator = operator
    
    def get_content(self):
        """
        A dictionary representation of the query.
        This dictionary is the content of the ``'query'`` key in the 
        JSON query.

        Returns
        -------
        content : dict
            The dictionary representation of the query.
        """
        content = {
            "type": "group",
            "logical_operator": self._operator,
            "nodes": [query.get_content() for query in self._queries]
        }
        return content



class BasicQuery(SingleQuery):
    """
    A text query for searching for a given term across all available
    fields.

    Parameters
    ----------
    term : str
        The search term.
        If the term contains multiple words, the query will return
        results where the entire term is present.
        The matching is not case-sensitive.
    
    Examples
    --------
    
    >>> query = BasicQuery("tc5b")
    >>> print(search(query))
    ['1L2Y']
    """
    def __init__(self, term):
        super().__init__()
        self._term = term

    def get_content(self):
        content = super().get_content()
        content["type"] = "terminal"
        content["service"] = "full_text"
        content["parameters"]["value"] = f'"{self._term}"'
        return content


class FieldQuery(SingleQuery):
    """
    A text query for searching for values in a given field using the
    given operator.

    The operators are keyword arguments of this function and the search
    value is the value given to the respective parameter.
    The operators are mutually exclusive.
    If none is given, the search will return results where the given
    field exists.

    A :class:`FieldQuery` is negated using the ``~`` operator.

    Parameters
    ----------
    field : str
        The field to search in.
    molecular_definition : bool, optional
        If set true, this query searches in fields
        associated with
        `molecular definitions <https://search.rcsb.org/chemical-search-attributes.html>`_.
        If false (default), this query searches in fields
        associated with `PDB structures <https://search.rcsb.org/structure-search-attributes.html>`_.
    exact_match : str, optional
        Operator for returning results whose field exactly matches the
        given value.
        The matching is not case-sensitive.
    contains_words, contains_phrase : str, optional
        Operator for returning results whose field matches
        individual words from the given value or the value as exact
        phrase, respectively.
        The matching is not case-sensitive.
    greater, less, greater_or_equal, less_or_equal, equals : int or float or datetime, optional
        Operator for returning results whose field values are larger,
        smaller or equal to the given value.
    range, range_closed : tuple(int, int) or tuple(float, float) or tuple(datetime, datetime), optional
        Operator for returning results whose field matches values within
        the given range. `range_closed` includes the interval limits.
    is_in : tuple of str or list of str, optional
        Operator for returning results whose field matches any of the
        values in the given list.

    Notes
    -----
    A complete list of the available fields and its supported operators
    is documented at
    `<https://search.rcsb.org/structure-search-attributes.html>`_
    and
    `<https://search.rcsb.org/chemical-search-attributes.html>`_.

    Examples
    --------
    
    >>> query = FieldQuery("reflns.d_resolution_high", less_or_equal=0.6)
    >>> print(sorted(search(query)))
    ['1EJG', '1I0T', '2GLT', '3NIR', '3P4J', '4JLJ', '5D8V', '5NW3', '7ATG']
    """
    def __init__(self, field, molecular_definition=False, **kwargs):
        super().__init__()
        self._negation = False
        self._field = field
        self._mol_definition = molecular_definition
        
        if len(kwargs) > 1:
            raise TypeError("Only one operator must be given")
        elif len(kwargs) == 1:
            self._operator = list(kwargs.keys())[0]
            self._value = list(kwargs.values())[0]
            if self._operator == "is_in":
                self._operator = "in"
        else:
            # No operator is given
            self._operator = "exists"
            self._value = None
        
        if self._operator not in [
            "exact_match",
            "contains_words", "contains_phrase",
            "greater", "less", "greater_or_equal", "less_or_equal", "equals",
            "range", "range_closed",
            "in",
            "exists"
        ]:
            raise TypeError(
                f"Constructor got an unexpected keyword argument "
                f"'{self._operator}'"
            )
        
        # Convert dates into ISO 8601
        if isinstance(self._value, datetime):
             self._value = _to_isoformat(self._value)
        elif isinstance(self._value, (tuple, list, np.ndarray)):
            self._value = [
                _to_isoformat(val) if isinstance(val, datetime) else val
                for val in self._value
            ]

    def get_content(self):
        content = super().get_content()
        content["type"] = "terminal"
        if self._mol_definition:
            content["service"] = "text_chem"
        else:
            content["service"] = "text"
        content["parameters"]["attribute"] = self._field
        content["parameters"]["operator"] = self._operator
        content["parameters"]["negation"] = self._negation
        if self._value is not None:
            content["parameters"]["value"] = self._value
        return content

    def __invert__(self):
        clone = copy.deepcopy(self)
        clone._negation = True
        return clone


class SequenceQuery(SingleQuery):
    """
    A query for protein/DNA/RNA molecules with a sequence similar to a
    given input sequence using
    `MMseqs2 <https://github.com/soedinglab/mmseqs2>`_.

    Parameters
    ----------
    sequence : Sequence or str
        The input sequence.
        If `sequence` is a :class:`NucleotideSequence` and the `scope`
        is ``'rna'``, ``'T'`` is automatically replaced by ``'U'``.
    scope : {'protein', 'dna', 'rna'}
        The type of molecule to find.
    min_identity : float, optional
        A match is only returned, if the sequence identity between
        the match and the input sequence exceeds this value.
        Must be between 0 and 1.
        By default, the sequence identity is ignored.
    max_expect_value : float, optional
        A match is only returned, if the *expect value* (E-value) does
        not exceed this value.
        By default, the value is effectively ignored.

    Notes
    -----
    *MMseqs2* is run on the RCSB servers.

    Examples
    --------
    
    >>> sequence = "NLYIQWLKDGGPSSGRPPPS"
    >>> query = SequenceQuery(sequence, scope="protein", min_identity=0.8)
    >>> print(sorted(search(query)))
    ['1L2Y', '1RIJ', '2JOF', '2LDJ', '2LL5', '2MJ9', '3UC7', '3UC8']
    """
    def __init__(self, sequence, scope,
                 min_identity=0.0, max_expect_value=10000000.0):
        super().__init__()
        self._target = _scope_to_target.get(scope.lower())
        if self._target is None:
            raise ValueError(f"'{scope}' is an invalid scope")
        
        if isinstance(sequence, NucleotideSequence) and scope.lower() == "rna":
            self._sequence = str(sequence).replace("T", "U")
        else:
            self._sequence = str(sequence)
        
        self._min_identity = min_identity
        self._max_expect_value = max_expect_value

    def get_content(self):
        content = super().get_content()
        content["type"] = "terminal"
        content["service"] = "sequence"
        content["parameters"]["value"] = self._sequence
        content["parameters"]["target"] = self._target
        content["parameters"]["identity_cutoff"] = self._min_identity
        content["parameters"]["evalue_cutoff"] = self._max_expect_value
        return content


class MotifQuery(SingleQuery):
    """
    A query for protein/DNA/RNA molecules containing the given sequence
    motif.

    Parameters
    ----------
    pattern : str
        The sequence pattern.
    pattern_type : {'simple', 'prosite', 'regex'}
        The type of the pattern.
    scope : {'protein', 'dna', 'rna'}
        The type of molecule to find.
    
    Examples
    --------
    
    >>> query = MotifQuery(
    ...     "C-x(2,4)-C-x(3)-[LIVMFYWC]-x(8)-H-x(3,5)-H.",
    ...     "prosite",
    ...     "protein"
    ... )
    """
    def __init__(self, pattern, pattern_type, scope):
        super().__init__()
        self._pattern = pattern
        self._pattern_type = pattern_type
        self._target = _scope_to_target.get(scope.lower())

    def get_content(self):
        content = super().get_content()
        content["type"] = "terminal"
        content["service"] = "seqmotif"
        content["parameters"]["value"] = self._pattern
        content["parameters"]["pattern_type"] = self._pattern_type
        content["parameters"]["target"] = self._target
        return content


class StructureQuery(SingleQuery):
    """
    A query for protein/DNA/RNA molecules with structural similarity
    to the query structure.

    Either the chain or assembly ID of the query structure must be
    specified.

    Parameters
    ----------
    pdb_id : str
        The PDB ID of the query structure.
    chain : str, optional
        The chain ID (more exactly ``asym_id``) of the query structure.
    assembly : str, optional
        The assembly ID (``assembly_id``) of the query structure.
    strict : bool, optional
        If true, structure comparison is strict, otherwise it is
        relaxed.
    
    Examples
    --------

    >>> query = StructureQuery("1L2Y", chain="A")
    >>> print(sorted(search(query)))
    ['1L2Y', '1RIJ', '2JOF', '2LDJ', '2M7D', '7MQS']
    """
    def __init__(self, pdb_id, chain=None, assembly=None, strict=True):
        super().__init__()

        if (chain is None and assembly is None) \
           or (chain is not None and assembly is not None):
                raise TypeError(
                    "Either the chain ID or assembly ID must be set"
                )
        elif chain is None:
            self._value = {
                "entry_id": pdb_id,
                "asssembly_id": assembly
            }
        else:
            self._value = {
                "entry_id": pdb_id,
                "asym_id": chain
            }
        
        self._operator = "strict_shape_match" if strict \
                         else "relaxed_shape_match"

    def get_content(self):
        content = super().get_content()
        content["type"] = "terminal"
        content["service"] = "structure"
        content["parameters"]["value"] = self._value
        content["parameters"]["operator"] = self._operator
        return content


def count(query, return_type="entry"):
    """
    Count PDB entries that meet the given query requirements,
    via the RCSB search API.
    
    This function requires an internet connection.
    
    Parameters
    ----------
    query : Query
        The search query.
    return_type : {'entry', 'assembly', 'polymer_entity', 'non_polymer_entity', 'polymer_instance'}, optional
        The type of the counted identifiers:

        - ``'entry'``: All macthing PDB entries are counted.
        - ``'assembly'``: All matching assemblies are counted.
        - ``'polymer_entity'``: All matching polymeric entities are
          counted.
        - ``'non_polymer_entity'``: All matching non-polymeric entities
          are counted.
        - ``'polymer_instance'``: All matching chains are counted.

    Returns
    -------
    ids : list of str
        A list of strings containing all PDB IDs that meet the query
        requirements.
    
    Examples
    --------
    
    >>> query = FieldQuery("reflns.d_resolution_high", less_or_equal=0.6)
    >>> print(count(query))
    9
    >>> ids = search(query)
    >>> print(sorted(ids))
    ['1EJG', '1I0T', '2GLT', '3NIR', '3P4J', '4JLJ', '5D8V', '5NW3', '7ATG']
    """
    if return_type not in [
        "entry", "polymer_instance", "assembly",
        "polymer_entity", "non_polymer_entity",
    ]:
        raise ValueError(f"'{return_type}' is an invalid return type")
    
    query_dict = {
        "query": query.get_content(),
        "return_type": return_type,
        "request_options": {
            # Do not return any IDs,
            # as we are only interested in the 'total_count' attribute
            "pager": {
                "start": 0,
                "rows": 0
            }
        }
    }
    r = requests.get(_search_url, params={"json": json.dumps(query_dict)})
    
    if r.status_code == 200:
        return r.json()["total_count"]
    elif r.status_code == 204:
        # Search did not return any results
        return 0
    else:
        try:
            raise RequestError(f"Error {r.status_code}: {r.json()['message']}")
        except json.decoder.JSONDecodeError:
            # In case there an error response without message
            raise RequestError(f"Error {r.status_code}")


def search(query, return_type="entry", range=None, sort_by=None):
    """
    Get all PDB IDs that meet the given query requirements,
    via the RCSB search API.
    
    This function requires an internet connection.
    
    Parameters
    ----------
    query : Query
        The search query.
    return_type : {'entry', 'assembly', 'polymer_entity', 'non_polymer_entity', 'polymer_instance'}, optional
        The type of the returned identifiers:

        - ``'entry'``: Only the PDB ID is returned (e.g. ``'XXXX'``).
          These can be used directly a input to :func:`fetch()`.
        - ``'assembly'``: The PDB ID appended with assembly ID is
          returned (e.g. ``'XXXX-1'``).
        - ``'polymer_entity'``: The PDB ID appended with entity ID of
          polymers is returned (e.g. ``'XXXX_1'``).
        - ``'non_polymer_entity'``: The PDB ID appended with entity ID
          of non-polymeric entities is returned (e.g. ``'XXXX_1'``).
        - ``'polymer_instance'``: The PDB ID appended with chain ID
          (more exactly ``'asym_id'``) is returned (e.g. ``'XXXX.A'``).
    
    range : tuple(int, int), optional
        If this parameter is specified, the PDB IDs in this range are
        selected from all matching PDB IDs and returned (pagination).
    sort_by : str, optional
        If specified, the returned PDB IDs are sorted by the values
        of the given field name in descending order.
        A complete list of the available fields is documented at
        `<https://search.rcsb.org/structure-search-attributes.html>`_.
        and
        `<https://search.rcsb.org/chemical-search-attributes.html>`_.

    Returns
    -------
    ids : list of str
        A list of strings containing all PDB IDs that meet the query
        requirements.

    Examples
    --------
    
    >>> query = FieldQuery("reflns.d_resolution_high", less_or_equal=0.6)
    >>> print(sorted(search(query)))
    ['1EJG', '1I0T', '2GLT', '3NIR', '3P4J', '4JLJ', '5D8V', '5NW3', '7ATG']
    >>> print(search(query, sort_by="rcsb_accession_info.initial_release_date"))
    ['7ATG', '5NW3', '5D8V', '4JLJ', '3P4J', '3NIR', '1I0T', '1EJG', '2GLT']
    >>> print(search(
    ...     query, range=(1,4), sort_by="rcsb_accession_info.initial_release_date"
    ... ))
    ['5NW3', '5D8V', '4JLJ']
    >>> print(sorted(search(query, return_type="polymer_instance")))
    ['1EJG.A', '1I0T.A', '1I0T.B', '2GLT.A', '3NIR.A', '3P4J.A', '3P4J.B', '4JLJ.A', '4JLJ.B', '5D8V.A', '5NW3.A', '7ATG.A', '7ATG.B']
    """
    if return_type not in [
        "entry", "polymer_instance", "assembly",
        "polymer_entity", "non_polymer_entity",
    ]:
        raise ValueError(f"'{return_type}' is an invalid return type")
    
    if sort_by is None:
        sort_by = "score"

    if range is None:
        start = 0
        rows = count(query, return_type=return_type)
    elif range[1] <= range[0]:
        raise ValueError("Range stop must be greater than range start")
    else:
        start = range[0]
        rows = range[1] - start

    query_dict = {
        "query": query.get_content(),
        "return_type": return_type,
        "request_options": {
            "pager": {
                "start": start,
                "rows": rows
            },
            "sort": [
                {
                    "sort_by": sort_by,
                }
            ]
        }
    }
    r = requests.get(_search_url, params={"json": json.dumps(query_dict)})
    
    if r.status_code == 200:
        return [result["identifier"] for result in r.json()["result_set"]]
    elif r.status_code == 204:
        # Search did not return any results
        return []
    else:
        try:
            raise RequestError(f"Error {r.status_code}: {r.json()['message']}")
        except json.decoder.JSONDecodeError:
            # In case there an error response without message
            raise RequestError(f"Error {r.status_code}")


def _to_isoformat(object):
    """
    Convert a datetime into the specifc ISO 8601 format required by the RCSB.
    """
    return object.strftime("%Y-%m-%dT%H:%M:%SZ")


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align


def test_from_alignment():
    seq1 = seq.NucleotideSequence("CGTCAT")
    seq2 = seq.NucleotideSequence("TCATGC")
    ali_str = ["CGTCAT--",
               "--TCATGC"]
    trace = align.Alignment.trace_from_strings(ali_str)
    alignment = align.Alignment([seq1, seq2], trace, None)

    profile = seq.SequenceProfile.from_alignment(alignment)
    symbols = np.array([[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 2], [0, 2, 0, 0],
                        [2, 0, 0, 0], [0, 0, 0, 2], [0, 0, 1, 0], [0, 1, 0, 0]])
    gaps = np.array([1, 1, 0, 0, 0, 0, 1, 1])
    alphabet = seq.Alphabet(["A", "C", "G", "T"])
    assert np.array_equal(symbols, profile.symbols)
    assert np.array_equal(gaps, profile.gaps)
    assert (alphabet == profile.alphabet)


def test_to_consensus_nuc():
    symbols = np.array([[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 2], [0, 2, 0, 0],
                        [2, 0, 0, 0], [0, 0, 0, 2], [0, 0, 1, 0], [0, 1, 0, 0]])
    gaps = np.array([1, 1, 0, 0, 0, 0, 1, 1])
    alphabet = seq.Alphabet(["A", "C", "G", "T"])
    profile = seq.SequenceProfile(symbols, gaps, alphabet)

    assert seq.NucleotideSequence("CGTCATGC") == profile.to_consensus()


def test_to_consensus_nuc_ambiguous():
    symbols = np.array([[1, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 2], [0, 2, 0, 0],
                        [2, 0, 0, 0], [0, 0, 0, 2], [0, 0, 1, 0], [0, 1, 0, 0]])
    gaps = np.array([1, 1, 0, 0, 0, 0, 1, 1])
    alphabet = seq.Alphabet(["A", "C", "G", "T"])
    profile = seq.SequenceProfile(symbols, gaps, alphabet)

    assert seq.NucleotideSequence("MGTCATGC") == profile.to_consensus()


def test_to_consensus_prot():
    # Avidin protein sequence
    seq1 = seq.ProteinSequence("MVHATSPLLLLLLLSLALVAPGLSARKCSLTGKWTNDLGSNMTIGAVNSRGEFTGTYITAVTATSNEIKESPLHGTQNTINKRTQP"
                               "TFGFTVNWKFSESTTVFTGQCFIDRNGKEVLKTMWLLRSSVNDIGDDWKATRVGINIFTRLRTQKE")
    # Streptavidin protein sequence
    seq2 = seq.ProteinSequence("MRKIVVAAIAVSLTTVSITASASADPSKDSKAQVSAAEAGITGTWYNQLGSTFIVTAGADGALTGTYESAVGNAESRYVLTGRYDSA"
                               "PATDGSGTALGWTVAWKNNYRNAHSATTWSGQYVGGAEARINTQWLLTSGTTEANAWKSTLVGHDTFTKVKPSAASIDAAKKAGVNN"
                               "GNPLDAVQQ")
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    alignment = align.align_optimal(seq1, seq2, matrix)[0]

    profile = seq.SequenceProfile.from_alignment(alignment)
    assert seq.ProteinSequence("MRHIATAAIALSLLLLSITALASADPGKDSKAQLSAAEAGITGKWTNDLGSNFIIGAVGADGAFTGTYESAVGNAESNEIKEGPLD"
                               "GAPATDGKGTALGWTFAFKNNWKFAESATTFSGQCFGGADARINGKELLTKGTMEANAWKSTLLGHDSFSKVKDIAADIDAAKKAG"
                               "INIFNPLDAQKE") == profile.to_consensus()


def test_new_position_matrices():
    seqs = [seq.NucleotideSequence("AAGAAT"),
            seq.NucleotideSequence("ATCATA"),
            seq.NucleotideSequence("AAGTAA"),
            seq.NucleotideSequence("AACAAA"),
            seq.NucleotideSequence("ATTAAA"),
            seq.NucleotideSequence("AAGAAT")]

    alignment = align.Alignment(
        sequences=seqs,
        trace=np.tile(np.arange(len(seqs[0])), len(seqs)) \
            .reshape(len(seqs), len(seqs[0])) \
            .transpose(),
        score=0
    )

    profile = seq.SequenceProfile.from_alignment(alignment)

    probability_matrix = np.array([[1., 0., 0., 0., ],
                                   [0.66666667, 0., 0., 0.33333333],
                                   [0., 0.33333333, 0.5, 0.16666667],
                                   [0.83333333, 0., 0., 0.16666667],
                                   [0.83333333, 0., 0., 0.16666667],
                                   [0.66666667, 0., 0., 0.33333333]])

    ppm = profile.probability_matrix()

    assert np.allclose(probability_matrix, ppm, atol=1e-3)

    probability = profile.sequence_probability(seq.NucleotideSequence("AAAAAA"))

    assert probability == 0.0

    ppm = profile.probability_matrix(pseudocount=1)

    probability_matrix = np.array([[0.89285714, 0.03571429, 0.03571429, 0.03571429],
                                   [0.60714286, 0.03571429, 0.03571429, 0.32142857],
                                   [0.03571429, 0.32142857, 0.46428571, 0.17857143],
                                   [0.75, 0.03571429, 0.03571429, 0.17857143],
                                   [0.75, 0.03571429, 0.03571429, 0.17857143],
                                   [0.60714286, 0.03571429, 0.03571429, 0.32142857]])

    assert np.allclose(probability_matrix, ppm, atol=1e-3)

    probability = profile.sequence_probability(seq.NucleotideSequence("AAAAAA"), pseudocount=1)

    assert probability == pytest.approx(0.0066, abs=1e-3)

    log_odds_matrix = np.array([[1.83650127, -2.80735492, -2.80735492, -2.80735492],
                                [1.28010792, -2.80735492, -2.80735492, 0.36257008],
                                [-2.80735492, 0.36257008, 0.8930848, -0.48542683],
                                [1.5849625, -2.80735492, -2.80735492, -0.48542683],
                                [1.5849625, -2.80735492, -2.80735492, -0.48542683],
                                [1.28010792, -2.80735492, -2.80735492, 0.36257008]])

    pwm = profile.log_odds_matrix(pseudocount=1)

    assert np.allclose(log_odds_matrix, pwm, atol=1e-3)

    score = profile.sequence_score(seq.NucleotideSequence("AAAAAA"), pseudocount=1)

    assert score == pytest.approx(4.7593, abs=1e-3)



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
The module contains the :class:`Sequence` superclass and :class:`GeneralSequence`.
"""

__name__ = "biotite.sequence"
__author__ = "Patrick Kunzmann"
__all__ = ["Sequence"]

import numbers
import abc
import numpy as np
from .alphabet import Alphabet, LetterAlphabet
from ..copyable import Copyable


_size_uint8  = np.iinfo(np.uint8 ).max +1
_size_uint16 = np.iinfo(np.uint16).max +1
_size_uint32 = np.iinfo(np.uint32).max +1


class Sequence(Copyable, metaclass=abc.ABCMeta):
    """
    The abstract base class for all sequence types.
    
    A :class:`Sequence` can be seen as a succession of symbols, that are
    elements in the allowed set of symbols, the :class:`Alphabet`.
    Internally, a :class:`Sequence` object uses a *NumPy*
    :class:`ndarray` of integers, where each integer represents a
    symbol.
    The :class:`Alphabet` of a :class:`Sequence` object is used to
    encode each symbol, that is used to create the
    :class:`Sequence`, into an integer. These integer values are called
    symbol code, the encoding of an entire sequence of symbols is
    called sequence code.
    
    The size of the symbol code type in the array is determined by the 
    size of the :class:`Alphabet`:
    If the :class:`Alphabet` contains 256 symbols or less, one byte is
    used per array element; if the :class:`Alphabet` contains
    between 257 and 65536 symbols, two bytes are used, and so on.
    
    Two :class:`Sequence` objects are equal if they are instances of the
    same class, have the same :class:`Alphabet` and have equal sequence
    codes.
    Comparison with a string or list of symbols evaluates always to
    false.
    
    A :class:`Sequence` can be indexed by any 1-D index a
    :class:`ndarray` accepts.
    If the index is a single integer, the decoded symbol at that
    position is returned, otherwise a subsequence is returned.
    
    Individual symbols of the sequence can also be exchanged in indexed
    form: If the an integer is used as index, the item is treated as a
    symbol. Any other index (slice, index list, boolean mask) expects
    multiple symbols, either as list of symbols, as :class:`ndarray`
    containing a sequence code or another :class:`Sequence` instance.
    Concatenation of two sequences is achieved with the '+' operator.
    
    Each subclass of :class:`Sequence` needs to overwrite the abstract
    method :func:`get_alphabet()`, which specifies the alphabet the
    :class:`Sequence` uses.
    
    Parameters
    ----------
    sequence : iterable object, optional
        The symbol sequence, the :class:`Sequence` is initialized with.
        For alphabets containing single letter strings, this parameter
        may also be a :class`str` object.
        By default the sequence is empty.
    
    Attributes
    ----------
    code : ndarray
        The sequence code.
    symbols : list
        The list of symbols, represented by the sequence.
        The list is generated by decoding the sequence code, when
        this attribute is accessed. When this attribute is modified,
        the new list of symbols is encoded into the sequence code.
    alphabet : Alphabet
        The alphabet of this sequence. Cannot be set.
        Equal to `get_alphabet()`.
    
    Examples
    --------
    Creating a DNA sequence from string and print the symbols and the
    code:
    
    >>> dna_seq = NucleotideSequence("ACGTA")
    >>> print(dna_seq)
    ACGTA
    >>> print(dna_seq.code)
    [0 1 2 3 0]
    >>> print(dna_seq.symbols)
    ['A' 'C' 'G' 'T' 'A']
    >>> print(list(dna_seq))
    ['A', 'C', 'G', 'T', 'A']
    
    Sequence indexing:
        
    >>> print(dna_seq[1:3])
    CG
    >>> print(dna_seq[[0,2,4]])
    AGA
    >>> print(dna_seq[np.array([False,False,True,True,True])])
    GTA
    
    Sequence manipulation:
        
    >>> dna_copy = dna_seq.copy()
    >>> dna_copy[2] = "C"
    >>> print(dna_copy)
    ACCTA
    >>> dna_copy = dna_seq.copy()
    >>> dna_copy[0:2] = dna_copy[3:5]
    >>> print(dna_copy)
    TAGTA
    >>> dna_copy = dna_seq.copy()
    >>> dna_copy[np.array([True,False,False,False,True])] = "T"
    >>> print(dna_copy)
    TCGTT
    >>> dna_copy = dna_seq.copy()
    >>> dna_copy[1:4] = np.array([0,1,2])
    >>> print(dna_copy)
    AACGA

    Reverse sequence:

    >>> dna_seq_rev = dna_seq.reverse()
    >>> print(dna_seq_rev)
    ATGCA
    
    Concatenate the two sequences:
        
    >>> dna_seq_concat = dna_seq + dna_seq_rev
    >>> print(dna_seq_concat)
    ACGTAATGCA
        
    """
    
    def __init__(self, sequence=()):
        self.symbols = sequence

    def copy(self, new_seq_code=None):
        """
        Copy the object.
        
        Parameters
        ----------
        new_seq_code : ndarray, optional
            If this parameter is set, the sequence code is set to this
            value, rather than the original sequence code.
        
        Returns
        -------
        copy
            A copy of this object.
        """
        # Override in order to achieve better performance,
        # in case only a subsequence is needed,
        # because not the entire sequence code is copied then
        clone = self.__copy_create__()
        if new_seq_code is None:
            clone.code = np.copy(self.code)
        else:
            clone.code = new_seq_code
        self.__copy_fill__(clone)
        return clone
    
    @property
    def symbols(self):
        return self.get_alphabet().decode_multiple(self.code)
    
    @symbols.setter
    def symbols(self, value):
        alph = self.get_alphabet()
        dtype = Sequence.dtype(len(alph))
        self._seq_code = alph.encode_multiple(value, dtype)
    
    @property
    def code(self):
        return self._seq_code
    
    @code.setter
    def code(self, value):
        dtype = Sequence.dtype(len(self.get_alphabet()))
        if not isinstance(value, np.ndarray):
            raise TypeError("Sequence code must be an integer ndarray")
        self._seq_code = value.astype(dtype, copy=False)
    
    @property
    def alphabet(self):
        return self.get_alphabet()
    
    @abc.abstractmethod
    def get_alphabet(self):
        """
        Get the :class:`Alphabet` of the :class:`Sequence`.
        
        This method must be overwritten, when subclassing
        :class:`Sequence`.
        
        Returns
        -------
        alphabet : Alphabet
            :class:`Sequence` alphabet.
        """
        pass
    
    def reverse(self, copy=True):
        """
        Reverse the :class:`Sequence`.
        
        Parameters
        ----------
        copy : bool, optional
            If set to False, the code :class:`ndarray` of the returned
            sequence is an array view to the sequence code of this
            object.
            In this case, manipulations on the returned sequence would
            also affect this object.
            Otherwise, the sequence code is copied.
        
        Returns
        -------
        reversed : Sequence
            The reversed :class:`Sequence`.
            
        Examples
        --------
            
        >>> dna_seq = NucleotideSequence("ACGTA")
        >>> dna_seq_rev = dna_seq.reverse()
        >>> print(dna_seq_rev)
        ATGCA
        """
        reversed_code = np.flip(self._seq_code, axis=0)
        if copy:
            reversed_code = np.copy(reversed_code)
        return self.copy(reversed_code)
    
    def is_valid(self):
        """
        Check, if the sequence contains a valid sequence code.
        
        A sequence code is valid, if at each sequence position the
        code is smaller than the size of the alphabet.
        
        Invalid code means that the code cannot be decoded into
        symbols. Furthermore invalid code can lead to serious
        errors in alignments, since the substitution matrix
        is indexed with an invalid index.
        
        Returns
        -------
        valid : bool
            True, if the sequence is valid, false otherwise.
        """
        return (self.code < len(self.get_alphabet())).all()
    
    def get_symbol_frequency(self):
        """
        Get the number of occurences of each symbol in the sequence.
        
        If a symbol does not occur in the sequence, but it is in the
        alphabet, its number of occurences is 0.
        
        Returns
        -------
        frequency : dict
            A dictionary containing the symbols as keys and the
            corresponding number of occurences in the sequence as
            values.
        """
        counts = np.bincount(
            self._seq_code, minlength=len(self.get_alphabet())
        )
        return {
            symbol: count for symbol, count
            in zip(self.get_alphabet().get_symbols(), counts)
        }
    
    def __getitem__(self, index):
        alph = self.get_alphabet()
        sub_seq = self._seq_code.__getitem__(index)
        if isinstance(sub_seq, np.ndarray):
            return self.copy(sub_seq)
        else:
            return alph.decode(sub_seq)
    
    def __setitem__(self, index, item):
        alph = self.get_alphabet()
        if isinstance(index, numbers.Integral):
            # Expect a single symbol
            code = alph.encode(item)
        else:
            # Expect multiple symbols
            if isinstance(item, Sequence):
                code = item.code
            elif isinstance(item, np.ndarray):
                code = item
            else:
                # Default: item is iterable object of symbols
                code = alph.encode_multiple(item)
        self._seq_code.__setitem__(index, code)
    
    def __len__(self):
        return len(self._seq_code)
    
    def __iter__(self):
        alph = self.get_alphabet()
        i = 0
        while i < len(self):
            yield alph.decode(self._seq_code[i])
            i += 1
    
    def __eq__(self, item):
        if not isinstance(item, type(self)):
            return False
        if self.get_alphabet() != item.get_alphabet():
            return False
        return np.array_equal(self._seq_code, item._seq_code)
    
    def __str__(self):
        alph = self.get_alphabet()
        if isinstance(alph, LetterAlphabet):
            return alph.decode_multiple(self._seq_code, as_bytes=True)\
                   .tobytes().decode("ASCII")
        else:
            return "".join(alph.decode_multiple(self._seq_code))
    
    def __add__(self, sequence):
        if self.get_alphabet().extends(sequence.get_alphabet()):
            new_code = np.concatenate((self._seq_code, sequence._seq_code))
            new_seq = self.copy(new_code)
            return new_seq
        elif sequence.get_alphabet().extends(self.get_alphabet()):
            new_code = np.concatenate((self._seq_code, sequence._seq_code))
            new_seq = sequence.copy(new_code)
            return new_seq
        else:
            raise ValueError("The sequences alphabets are not compatible")

    @staticmethod
    def dtype(alphabet_size):
        """
        Get the sequence code dtype required for the given size of the
        alphabet.

        Parameters
        ----------
        alpahabet_size : int
            The size of the alphabet.
        
        Returns
        -------
        dtype
            The  :class:`dtype`, that is large enough to store symbol
            codes, that are encoded by an :class:`Alphabet` of the given
            size.
        """
        if alphabet_size <= _size_uint8:
            return np.uint8
        elif alphabet_size <= _size_uint16:
            return np.uint16
        elif alphabet_size <= _size_uint32:
            return np.uint32
        else:
            return np.uint64



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pytest
import numpy as np
import biotite.sequence as seq


def test_encoding():
    string1 = "AATGCGTTA"
    dna = seq.NucleotideSequence(string1)
    string2 = str(dna)
    assert string1 == string2

def test_validity_check():
    dna = seq.NucleotideSequence()
    dna.code = np.array([0,1,0,3,3])
    assert dna.is_valid()
    dna.code = np.array([0,1,4,3,3])
    assert not dna.is_valid()
    
def test_access():
    string = "AATGCGTTA"
    dna = seq.NucleotideSequence(string)
    assert string[2] == dna[2]
    assert string == "".join([symbol for symbol in dna])
    dna = dna[3:-2]
    assert "GCGT" == str(dna)

def test_manipulation():
    dna_seq = seq.NucleotideSequence("ACGTA")
    dna_copy = dna_seq.copy()
    dna_copy[2] = "C"
    assert "ACCTA" == str(dna_copy)
    dna_copy = dna_seq.copy()
    dna_copy[0:2] = dna_copy[3:5]
    assert "TAGTA" == str(dna_copy)
    dna_copy = dna_seq.copy()
    dna_copy[np.array([True,False,False,False,True])] = "T"
    assert "TCGTT" == str(dna_copy)
    dna_copy = dna_seq.copy()
    dna_copy[1:4] = np.array([0,1,2])
    assert "AACGA" == str(dna_copy)

def test_concatenation():
    str1 = "AAGTTA"
    str2 = "CGA"
    str3 = "NNN"
    concat_seq = seq.NucleotideSequence(str1) + seq.NucleotideSequence(str2)
    assert str1 + str2 == str(concat_seq)
    concat_seq = seq.NucleotideSequence(str1) + seq.NucleotideSequence(str3)
    assert str1 + str3 == str(concat_seq)
    concat_seq = seq.NucleotideSequence(str3) + seq.NucleotideSequence(str1)
    assert str3 + str1 == str(concat_seq)

def test_frequency():
    string = "ACGCGAGAAAGCGGG"
    dna = seq.NucleotideSequence(string)
    assert dna.get_symbol_frequency() == {"A": 5, "C": 3, "G": 7, "T": 0}
    
def test_alph_error():
    string = "AATGCGTUTA"
    with pytest.raises(seq.AlphabetError):
        seq.NucleotideSequence(string)

def test_alphabet_extension():
    alph1 = seq.Alphabet("abc")
    alph2 = seq.Alphabet("abc")
    alph3 = seq.Alphabet("acb")
    alph4 = seq.Alphabet("abcde")
    assert alph1.extends(alph1)
    assert alph2.extends(alph1)
    assert not alph3.extends(alph1)
    assert alph4.extends(alph1)
    assert not alph1.extends(alph4)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functions to calculate atomistic densities.
"""

__name__ = "biotite.structure"
__author__ = "Daniel Bauer"
__all__ = ["density"]

import numpy as np
from .atoms import coord


def density(atoms, selection=None, delta=1.0, bins=None,
            density=False, weights=None):
    r"""
    Compute the density of the selected atoms.

    This creates a 3d histogram over the coordinates of selected atoms.
    By default, the grid for the histogram is built based on the
    coordinates of the given `atoms` with an even gridspacing of
    `delta` in all three dimensions.
    Alternatively, a custom grid can be used.

    Parameters
    ----------
    atoms : AtomArray or AtomArrayStack or ndarray, shape=(n,3) or shape=(m,n,3)
        The density is calculated based on these atoms.
        Alternatively, the coordinates can be directly provided as
        `ndarray`.
    selection : ndarray, dtype=bool, shape=(n,), optional
        Boolean mask for `atoms` to calculate the density only on a set
        of atoms.
    delta : float, optional
        Distance between grid points for density calculation (in Å).
    bins : int or sequence of scalars or str, optional
        Bins for the RDF.

        - If `bins` is an `int`, it defines the number of bins.
        - If `bins` is a sequence, it defines the bin edges, ignoring
          the actual coordinates of the `atoms` selection.
        - If `bins` is a string, it defines the function used to
          calculate the bins.

        See :func:`numpy.histogramdd()` for further details.
    density : boolean, optional
        If False, the number of samples in each bin is returned.
        Otherwise, returns the probability density function of each bin.
        See :func:`numpy.histogramdd()` for further details.
    weights: ndarray, shape=(n,) or shape=(m,n), optional
        An array of values to weight the contribution of *n* atoms in 
        *m* models.
        If the shape is *(n,)*, the weights will be interpreted as
        *per atom*.
        A shape of *(m,n)* allows to additionally weight atoms on a
        *per model* basis.
    
    Returns
    -------
    H : ndarray, dtype=float
        The threedimensional histogram of the selected atoms.
        The histogram takes the atoms in all models into account.
        The length of the histogram depends on `atoms` coordinates and
        `delta`, or the supplied `bins` input parameter.
    edges : list of ndarray, dtype=float
        A list containing the 3 arrays describing the bin edges.
    """
    coords = coord(atoms)
    
    is_stack = coords.ndim == 3

    # Define the grid for coordinate binning based on coordinates of
    # supplied atoms
    # This makes the binning independent of a supplied box vector and 
    # fluctuating box dimensions are not a problem
    # However, this means that the user has to make sure the region of
    # interest is in the center of the box, i.e. by centering the
    # investigated protein in the box.
    if bins is None:
        if is_stack:
            axis = (0, 1)
        else:
            axis = 0
        grid_min, grid_max = np.min(
            coords, axis=axis), np.max(coords, axis=axis
        )
        bins = [
            np.arange(grid_min[0], grid_max[0]+delta, delta),
            np.arange(grid_min[1], grid_max[1]+delta, delta),
            np.arange(grid_min[2], grid_max[2]+delta, delta),
        ]

    if selection is None:
        selected_coords = coords
    else:
        selected_coords = coords[...,selection, :]

    # Reshape the coords into Nx3
    coords = selected_coords.reshape((np.prod(selected_coords.shape[:-1]), 3))

    # We need a weight value per coordinate, but input might be per atom
    if weights is not None:
        if is_stack and len(weights.shape) < 2:
            weights = np.tile(weights, len(selected_coords))
        weights = weights.reshape(coords.shape[0])
    
    # Calculate the histogram
    hist = np.histogramdd(
        coords, bins=bins, density=density, weights=weights
    )
    return hist
